# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import glob
import os
import re
import time
from typing import Optional, List, Tuple, Set

from gi.repository import GLib

from inputremapper.configs.global_config import GlobalConfig
from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import UIMapping, MappingData
from inputremapper.configs.paths import PathUtils
from inputremapper.configs.preset import Preset
from inputremapper.configs.keyboard_layout import KeyboardLayout
from inputremapper.daemon import DaemonProxy
from inputremapper.exceptions import DataManagementError
from inputremapper.groups import _Group
from inputremapper.gui.gettext import _
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
)
from inputremapper.gui.messages.message_data import (
    UInputsData,
    GroupData,
    PresetData,
    CombinationUpdate,
)
from inputremapper.gui.reader_client import ReaderClient
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.injector import (
    InjectorState,
    InjectorStateMessage,
)
from inputremapper.logging.logger import logger

DEFAULT_PRESET_NAME = _("new preset")

# useful type aliases
Name = str
GroupKey = str


class DataManager:
    """DataManager provides an interface to create and modify configurations as well
    as modify the state of the Service.

    Any state changes will be announced via the MessageBroker.
    """

    def __init__(
        self,
        message_broker: MessageBroker,
        config: GlobalConfig,
        reader_client: ReaderClient,
        daemon: DaemonProxy,
        uinputs: GlobalUInputs,
        keyboard_layout: KeyboardLayout,
    ):
        self.message_broker = message_broker
        self._reader_client = reader_client
        self._daemon = daemon
        self._uinputs = uinputs
        self._keyboard_layout = keyboard_layout
        uinputs.prepare_all()

        self._config = config
        self._config.load_config()

        self._active_preset: Optional[Preset[UIMapping]] = None
        self._active_mapping: Optional[UIMapping] = None
        self._active_input_config: Optional[InputConfig] = None

    def publish_group(self):
        """Send active group to the MessageBroker.

        This is internally called whenever the group changes.
        It is usually not necessary to call this explicitly from
        outside DataManager.
        """
        self.message_broker.publish(
            GroupData(self.active_group.key, self.get_preset_names())
        )

    def publish_preset(self):
        """Send active preset to the MessageBroker.

        This is internally called whenever the preset changes.
        It is usually not necessary to call this explicitly from
        outside DataManager.
        """
        self.message_broker.publish(
            PresetData(
                self.active_preset.name, self.get_mappings(), self.get_autoload()
            )
        )

    def publish_mapping(self):
        """Send active mapping to the MessageBroker

        This is internally called whenever the mapping changes.
        It is usually not necessary to call this explicitly from
        outside DataManager.
        """
        if self.active_mapping:
            self.message_broker.publish(self.active_mapping.get_bus_message())

    def publish_event(self):
        """Send active event to the MessageBroker.

        This is internally called whenever the event changes.
        It is usually not necessary to call this explicitly from
        outside DataManager
        """
        if self.active_input_config:
            assert self.active_input_config in self.active_mapping.input_combination
            self.message_broker.publish(self.active_input_config)

    def publish_uinputs(self):
        """Send the "uinputs" message on the MessageBroker."""
        self.message_broker.publish(
            UInputsData(
                {
                    name: uinput.capabilities()
                    for name, uinput in self._uinputs.devices.items()
                }
            )
        )

    def publish_groups(self):
        """Publish the "groups" message on the MessageBroker."""
        self._reader_client.publish_groups()

    def publish_injector_state(self):
        """Publish the "injector_state" message for the active_group."""
        if not self.active_group:
            return

        self.message_broker.publish(InjectorStateMessage(self.get_state()))

    @property
    def active_group(self) -> Optional[_Group]:
        """The currently loaded group."""
        return self._reader_client.group

    @property
    def active_preset(self) -> Optional[Preset[UIMapping]]:
        """The currently loaded preset."""
        return self._active_preset

    @property
    def active_mapping(self) -> Optional[UIMapping]:
        """The currently loaded mapping."""
        return self._active_mapping

    @property
    def active_input_config(self) -> Optional[InputConfig]:
        """The currently loaded event."""
        return self._active_input_config

    def get_group_keys(self) -> Tuple[GroupKey, ...]:
        """Get all group keys (plugged devices)."""
        return tuple(group.key for group in self._reader_client.groups.filter())

    def get_preset_names(self) -> Tuple[Name, ...]:
        """Get all preset names for active_group and current user sorted by age."""
        if not self.active_group:
            raise DataManagementError("Cannot find presets: Group is not set")
        device_folder = PathUtils.get_preset_path(self.active_group.name)
        PathUtils.mkdir(device_folder)

        paths = glob.glob(os.path.join(glob.escape(device_folder), "*.json"))
        presets = [
            os.path.splitext(os.path.basename(path))[0]
            for path in sorted(paths, key=os.path.getmtime)
        ]
        # the highest timestamp to the front
        presets.reverse()
        return tuple(presets)

    def get_mappings(self) -> Optional[List[MappingData]]:
        """All mappings from the active_preset."""
        if not self._active_preset:
            return None

        return [mapping.get_bus_message() for mapping in self._active_preset]

    def get_autoload(self) -> bool:
        """The autoload status of the active_preset."""
        if not self.active_preset or not self.active_group:
            return False
        return self._config.is_autoloaded(
            self.active_group.key, self.active_preset.name
        )

    def set_autoload(self, status: bool):
        """Set the autoload status of the active_preset.

        Will send "preset" message on the MessageBroker.
        """
        if not self.active_preset or not self.active_group:
            raise DataManagementError("Cannot set autoload status: Preset is not set")

        if status:
            self._config.set_autoload_preset(
                self.active_group.key, self.active_preset.name
            )
        elif self.get_autoload():
            self._config.set_autoload_preset(self.active_group.key, None)

        self.publish_preset()

    def get_newest_group_key(self) -> GroupKey:
        """group_key of the group with the most recently modified preset."""
        paths = []
        pattern = os.path.join(
            glob.escape(PathUtils.get_preset_path()),
            "*/*.json",
        )
        for path in glob.glob(pattern):
            if self._reader_client.groups.find(key=PathUtils.split_all(path)[-2]):
                paths.append((path, os.path.getmtime(path)))

        if not paths:
            raise FileNotFoundError()

        path, _ = max(paths, key=lambda x: x[1])
        return PathUtils.split_all(path)[-2]

    def get_newest_preset_name(self) -> Name:
        """Preset name of the most recently modified preset in the active group."""
        if not self.active_group:
            raise DataManagementError("Cannot find newest preset: Group is not set")

        pattern = os.path.join(
            glob.escape(PathUtils.get_preset_path(self.active_group.name)),
            "*.json",
        )
        paths = [(path, os.path.getmtime(path)) for path in glob.glob(pattern)]
        if not paths:
            raise FileNotFoundError()

        path, _ = max(paths, key=lambda x: x[1])
        return os.path.split(path)[-1].split(".")[0]

    def get_available_preset_name(self, name=DEFAULT_PRESET_NAME) -> Name:
        """The first available preset in the active group."""
        if not self.active_group:
            raise DataManagementError("Unable find preset name. Group is not set")

        name = name.strip()

        # find a name that is not already taken
        if os.path.exists(PathUtils.get_preset_path(self.active_group.name, name)):
            # if there already is a trailing number, increment it instead of
            # adding another one
            match = re.match(r"^(.+) (\d+)$", name)
            if match:
                name = match[1]
                i = int(match[2]) + 1
            else:
                i = 2

            while os.path.exists(
                PathUtils.get_preset_path(self.active_group.name, f"{name} {i}")
            ):
                i += 1

            return f"{name} {i}"

        return name

    def load_group(self, group_key: str):
        """Load a group. will publish "groups" and "injector_state" messages.

        This will render the active_mapping and active_preset invalid.
        """
        if group_key not in self.get_group_keys():
            raise DataManagementError("Unable to load non existing group")

        logger.info('Loading group "%s"', group_key)

        self._active_input_config = None
        self._active_mapping = None
        self._active_preset = None
        group = self._reader_client.groups.find(key=group_key)
        self._reader_client.set_group(group)
        self.publish_group()
        self.publish_injector_state()

    def load_preset(self, name: str):
        """Load a preset. Will send "preset" message on the MessageBroker.

        This will render the active_mapping invalid.
        """
        if not self.active_group:
            raise DataManagementError("Unable to load preset. Group is not set")

        logger.info('Loading preset "%s"', name)

        preset_path = PathUtils.get_preset_path(self.active_group.name, name)
        preset = Preset(preset_path, mapping_factory=UIMapping)
        preset.load()
        self._active_input_config = None
        self._active_mapping = None
        self._active_preset = preset
        self.publish_preset()

    def load_mapping(self, combination: InputCombination):
        """Load a mapping. Will send "mapping" message on the MessageBroker."""
        if not self._active_preset:
            raise DataManagementError("Unable to load mapping. Preset is not set")

        mapping = self._active_preset.get_mapping(combination)
        if not mapping:
            msg = (
                f"the mapping with {combination = } does not "
                f"exist in the {self._active_preset.path}"
            )
            logger.error(msg)
            raise KeyError(msg)

        self._active_input_config = None
        self._active_mapping = mapping
        self.publish_mapping()

    def load_input_config(self, input_config: InputConfig):
        """Load a InputConfig from the combination in the active mapping.

        Will send "event" message on the MessageBroker,
        """
        if not self.active_mapping:
            raise DataManagementError("Unable to load event. Mapping is not set")
        if input_config not in self.active_mapping.input_combination:
            raise ValueError(
                f"{input_config} is not member of active_mapping.input_combination: "
                f"{self.active_mapping.input_combination}"
            )
        self._active_input_config = input_config
        self.publish_event()

    def rename_preset(self, new_name: str):
        """Rename the current preset and move the correct file.

        Will send "group" and then "preset" message on the MessageBroker
        """
        if not self.active_preset or not self.active_group:
            raise DataManagementError("Unable rename preset: Preset is not set")

        if self.active_preset.path == PathUtils.get_preset_path(
            self.active_group.name, new_name
        ):
            return

        old_path = self.active_preset.path
        assert old_path is not None
        old_name = os.path.basename(old_path).split(".")[0]
        new_path = PathUtils.get_preset_path(self.active_group.name, new_name)
        if os.path.exists(new_path):
            raise ValueError(
                f"cannot rename {old_name} to " f"{new_name}, preset already exists"
            )

        logger.info('Moving "%s" to "%s"', old_path, new_path)
        os.rename(old_path, new_path)
        now = time.time()
        os.utime(new_path, (now, now))

        if self._config.is_autoloaded(self.active_group.key, old_name):
            self._config.set_autoload_preset(self.active_group.key, new_name)

        self.active_preset.path = PathUtils.get_preset_path(
            self.active_group.name, new_name
        )
        self.publish_group()
        self.publish_preset()

    def copy_preset(self, name: str):
        """Copy the current preset to the given name.

        Will send "group" and "preset" message to the MessageBroker and load the copy
        """
        # todo: Do we want to load the copy here? or is this up to the controller?
        if not self.active_preset or not self.active_group:
            raise DataManagementError("Unable to copy preset: Preset is not set")

        if self.active_preset.path == PathUtils.get_preset_path(
            self.active_group.name, name
        ):
            return

        if name in self.get_preset_names():
            raise ValueError(f"a preset with the name {name} already exits")

        new_path = PathUtils.get_preset_path(self.active_group.name, name)
        logger.info('Copy "%s" to "%s"', self.active_preset.path, new_path)
        self.active_preset.path = new_path
        self.save()
        self.publish_group()
        self.publish_preset()

    def create_preset(self, name: str):
        """Create empty preset in the active_group.

        Will send "group" message to the MessageBroker
        """
        if not self.active_group:
            raise DataManagementError("Unable to add preset. Group is not set")

        path = PathUtils.get_preset_path(self.active_group.name, name)
        if os.path.exists(path):
            raise DataManagementError("Unable to add preset. Preset exists")

        Preset(path).save()
        self.publish_group()

    def delete_preset(self):
        """Delete the active preset.

        Will send "group" message to the MessageBroker
        this will invalidate the active mapping,
        """
        preset_path = self._active_preset.path
        logger.info('Removing "%s"', preset_path)
        os.remove(preset_path)
        self._active_mapping = None
        self._active_preset = None
        self.publish_group()

    def update_mapping(self, **kwargs):
        """Update the active mapping with the given keywords and values.

        Will send "mapping" message to the MessageBroker. In case of a new
        input_combination. This will first send a "combination_update" message.
        """
        if not self._active_mapping:
            raise DataManagementError("Cannot modify Mapping: Mapping is not set")

        if symbol := kwargs.get("output_symbol"):
            kwargs["output_symbol"] = self._keyboard_layout.correct_case(symbol)

        combination = self.active_mapping.input_combination
        for key, value in kwargs.items():
            setattr(self._active_mapping, key, value)

        if (
            "input_combination" in kwargs
            and combination != self.active_mapping.input_combination
        ):
            self._active_input_config = None
            self.message_broker.publish(
                CombinationUpdate(combination, self._active_mapping.input_combination)
            )

        if "mapping_type" in kwargs:
            # mapping_type must be the last update because it is automatically updated
            # by a validation function
            self._active_mapping.mapping_type = kwargs["mapping_type"]

        self.publish_mapping()

    def update_input_config(self, new_input_config: InputConfig):
        """Update the active input configuration.

        Will send "combination_update", "mapping" and "event" messages to the
        MessageBroker (in that order)
        """
        if not self.active_mapping or not self.active_input_config:
            raise DataManagementError("Cannot modify event: Event is not set")

        combination = list(self.active_mapping.input_combination)
        combination[combination.index(self.active_input_config)] = new_input_config
        self.update_mapping(input_combination=InputCombination(combination))
        self._active_input_config = new_input_config
        self.publish_event()

    def create_mapping(self):
        """Create empty mapping in the active preset.

        Will send "preset" message to the MessageBroker
        """
        if not self._active_preset:
            raise DataManagementError("Cannot create mapping: Preset is not set")
        self._active_preset.add(UIMapping())
        self.publish_preset()

    def delete_mapping(self):
        """Delete the active mapping.

        Will send "preset" message to the MessageBroker
        """
        if not self._active_mapping:
            raise DataManagementError(
                "cannot delete active mapping: active mapping is not set"
            )

        self._active_preset.remove(self._active_mapping.input_combination)
        self._active_mapping = None
        self.publish_preset()

    def save(self):
        """Save the active preset."""
        if self._active_preset:
            self._active_preset.save()

    def refresh_groups(self):
        """Refresh the groups (plugged devices).

        Should send "groups" message to MessageBroker this will not happen immediately
        because the system might take a bit until the groups are available
        """
        self._reader_client.refresh_groups()

    def start_combination_recording(self):
        """Record user input.

        Will send "combination_recorded" messages as new input arrives.
        Will eventually send a "recording_finished" message.
        """
        self._reader_client.start_recorder()

    def stop_combination_recording(self):
        """Stop recording user input.

        Will send a recording_finished signal if a recording is running.
        """
        self._reader_client.stop_recorder()

    def stop_injecting(self) -> None:
        """Stop injecting for the active group.

        Will send "injector_state" message once the injector has stopped."""
        if not self.active_group:
            raise DataManagementError("Cannot stop injection: Group is not set")
        self._daemon.stop_injecting(self.active_group.key)
        self.do_when_injector_state(
            {InjectorState.STOPPED}, self.publish_injector_state
        )

    def start_injecting(self) -> bool:
        """Start injecting the active preset for the active group.

        returns if the startup was successfully initialized.
        Will send "injector_state" message once the startup is complete.
        """
        if not self.active_preset or not self.active_group:
            raise DataManagementError("Cannot start injection: Preset is not set")

        self._daemon.set_config_dir(self._config.get_dir())
        assert self.active_preset.name is not None
        if self._daemon.start_injecting(self.active_group.key, self.active_preset.name):
            self.do_when_injector_state(
                {
                    InjectorState.RUNNING,
                    InjectorState.ERROR,
                    InjectorState.NO_GRAB,
                    InjectorState.UPGRADE_EVDEV,
                },
                self.publish_injector_state,
            )
            return True
        return False

    def get_state(self) -> InjectorState:
        """The state of the injector."""
        if not self.active_group:
            raise DataManagementError("Cannot read state: Group is not set")
        return self._daemon.get_state(self.active_group.key)

    def refresh_service_config_path(self):
        """Tell the service to refresh its config path."""
        self._daemon.set_config_dir(self._config.get_dir())

    def do_when_injector_state(self, states: Set[InjectorState], callback):
        """Run callback once the injector state is one of states."""
        start = time.time()

        def do():
            if time.time() - start > 3:
                # something went wrong, there should have been a state long ago.
                # the timeout prevents tons of GLib.timeouts to run forever, especially
                # after spamming the "Stop" button.
                logger.error("Timed out while waiting for injector state %s", states)
                return False

            if self.get_state() in states:
                callback()
                return False
            return True

        GLib.timeout_add(100, do)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import unittest

from evdev.ecodes import (
    EV_KEY,
    KEY_A,
    KEY_B,
)

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import DummyMapping, MacroTestBase


@test_setup
class TestIfTap(MacroTestBase):
    async def test_if_tap(self):
        macro = Parser.parse("if_tap(key(x), key(y), 100)", self.context, DummyMapping)
        self.assertEqual(self.count_child_macros(macro), 2)

        x = keyboard_layout.get("x")
        y = keyboard_layout.get("y")

        # this is the regular routine of how a macro is started. the tigger is pressed
        # already when the macro runs, and released during if_tap within the timeout.
        macro.press_trigger()
        asyncio.ensure_future(macro.run(self.handler))
        await asyncio.sleep(0.05)
        macro.release_trigger()
        await asyncio.sleep(0.05)

        self.assertListEqual(self.result, [(EV_KEY, x, 1), (EV_KEY, x, 0)])
        self.assertFalse(macro.running)

    async def test_if_tap_2(self):
        # when the press arrives shortly after run.
        # a tap will happen within the timeout even if the tigger is not pressed when
        # it does into if_tap
        macro = Parser.parse("if_tap(key(a), key(b), 100)", self.context, DummyMapping)
        asyncio.ensure_future(macro.run(self.handler))

        await asyncio.sleep(0.01)
        macro.press_trigger()
        await asyncio.sleep(0.01)
        macro.release_trigger()
        await asyncio.sleep(0.2)
        self.assertListEqual(self.result, [(EV_KEY, KEY_A, 1), (EV_KEY, KEY_A, 0)])
        self.assertFalse(macro.running)
        self.result.clear()

    async def test_if_double_tap(self):
        macro = Parser.parse(
            "if_tap(if_tap(key(a), key(b), 100), key(c), 100)",
            self.context,
            DummyMapping,
        )
        self.assertEqual(self.count_child_macros(macro), 4)
        self.assertEqual(self.count_tasks(macro), 5)

        asyncio.ensure_future(macro.run(self.handler))

        # first tap
        macro.press_trigger()
        await asyncio.sleep(0.05)
        macro.release_trigger()

        # second tap
        await asyncio.sleep(0.04)
        macro.press_trigger()
        await asyncio.sleep(0.04)
        macro.release_trigger()

        await asyncio.sleep(0.05)
        self.assertListEqual(self.result, [(EV_KEY, KEY_A, 1), (EV_KEY, KEY_A, 0)])
        self.assertFalse(macro.running)
        self.result.clear()

        """If the second tap takes too long, runs else there"""

        asyncio.ensure_future(macro.run(self.handler))

        # first tap
        macro.press_trigger()
        await asyncio.sleep(0.05)
        macro.release_trigger()

        # second tap
        await asyncio.sleep(0.06)
        macro.press_trigger()
        await asyncio.sleep(0.06)
        macro.release_trigger()

        await asyncio.sleep(0.05)
        self.assertListEqual(self.result, [(EV_KEY, KEY_B, 1), (EV_KEY, KEY_B, 0)])
        self.assertFalse(macro.running)
        self.result.clear()

    async def test_if_tap_none(self):
        # first param none
        macro = Parser.parse("if_tap(, key(y), 100)", self.context, DummyMapping)
        self.assertEqual(self.count_child_macros(macro), 1)
        y = keyboard_layout.get("y")
        macro.press_trigger()
        asyncio.ensure_future(macro.run(self.handler))
        await asyncio.sleep(0.05)
        macro.release_trigger()
        await asyncio.sleep(0.05)
        self.assertListEqual(self.result, [])

        # second param none
        macro = Parser.parse("if_tap(key(y), , 50)", self.context, DummyMapping)
        self.assertEqual(self.count_child_macros(macro), 1)
        y = keyboard_layout.get("y")
        macro.press_trigger()
        asyncio.ensure_future(macro.run(self.handler))
        await asyncio.sleep(0.1)
        macro.release_trigger()
        await asyncio.sleep(0.05)
        self.assertListEqual(self.result, [])

        self.assertFalse(macro.running)

    async def test_if_not_tap(self):
        macro = Parser.parse("if_tap(key(x), key(y), 50)", self.context, DummyMapping)
        self.assertEqual(self.count_child_macros(macro), 2)

        y = keyboard_layout.get("y")

        macro.press_trigger()
        asyncio.ensure_future(macro.run(self.handler))
        await asyncio.sleep(0.1)
        macro.release_trigger()
        await asyncio.sleep(0.05)

        self.assertListEqual(self.result, [(EV_KEY, y, 1), (EV_KEY, y, 0)])
        self.assertFalse(macro.running)

    async def test_if_not_tap_named(self):
        macro = Parser.parse(
            "if_tap(key(x), key(y), timeout=50)", self.context, DummyMapping
        )
        self.assertEqual(self.count_child_macros(macro), 2)

        x = keyboard_layout.get("x")
        y = keyboard_layout.get("y")

        macro.press_trigger()
        asyncio.ensure_future(macro.run(self.handler))
        await asyncio.sleep(0.1)
        macro.release_trigger()
        await asyncio.sleep(0.05)

        self.assertListEqual(self.result, [(EV_KEY, y, 1), (EV_KEY, y, 0)])
        self.assertFalse(macro.running)

    async def test_raises_error(self):
        Parser.parse("if_tap(, k(a), 1000)", self.context)  # no error
        Parser.parse("if_tap(, k(a), timeout=1000)", self.context)  # no error
        Parser.parse("if_tap(, k(a), $timeout)", self.context)  # no error
        Parser.parse("if_tap(, k(a), timeout=$t)", self.context)  # no error
        Parser.parse("if_tap(, key(a))", self.context)  # no error
        Parser.parse("if_tap(k(a),)", self.context)  # no error
        self.assertRaises(MacroError, Parser.parse, "if_tap(k(a), b)", self.context)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import re
import unittest

from evdev.ecodes import EV_KEY, KEY_A, KEY_B, KEY_C, KEY_E

from inputremapper.configs.validation_errors import (
    MacroError,
)
from inputremapper.injection.macros.argument import Argument, ArgumentConfig
from inputremapper.injection.macros.parse import Parser
from inputremapper.injection.macros.raw_value import RawValue
from inputremapper.injection.macros.tasks.hold_keys import HoldKeysTask
from inputremapper.injection.macros.tasks.if_tap import IfTapTask
from inputremapper.injection.macros.tasks.key import KeyTask
from inputremapper.injection.macros.variable import Variable
from tests.lib.logger import logger
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import DummyMapping, MacroTestBase


@test_setup
class TestParsing(MacroTestBase):
    def test_get_macro_argument_names(self):
        self.assertEqual(
            IfTapTask.get_macro_argument_names(),
            ["then", "else", "timeout"],
        )

        self.assertEqual(
            HoldKeysTask.get_macro_argument_names(),
            ["*symbols"],
        )

    def test_get_num_parameters(self):
        self.assertEqual(IfTapTask.get_num_parameters(), (0, 3))
        self.assertEqual(KeyTask.get_num_parameters(), (1, 1))
        self.assertEqual(HoldKeysTask.get_num_parameters(), (0, float("inf")))

    def test_remove_whitespaces(self):
        self.assertEqual(Parser.remove_whitespaces('foo"bar"foo'), 'foo"bar"foo')
        self.assertEqual(Parser.remove_whitespaces('foo" bar"foo'), 'foo" bar"foo')
        self.assertEqual(
            Parser.remove_whitespaces('foo" bar"fo" "o'), 'foo" bar"fo" "o'
        )
        self.assertEqual(
            Parser.remove_whitespaces(' fo o"\nba r "f\noo'), 'foo"\nba r "foo'
        )
        self.assertEqual(Parser.remove_whitespaces(' a " b " c " '), 'a" b "c" ')

        self.assertEqual(Parser.remove_whitespaces('"""""""""'), '"""""""""')
        self.assertEqual(Parser.remove_whitespaces('""""""""'), '""""""""')

        self.assertEqual(Parser.remove_whitespaces("      "), "")
        self.assertEqual(Parser.remove_whitespaces('     " '), '" ')
        self.assertEqual(Parser.remove_whitespaces('     " " '), '" "')

        self.assertEqual(Parser.remove_whitespaces("a# ##b", delimiter="##"), "a###b")
        self.assertEqual(Parser.remove_whitespaces("a###b", delimiter="##"), "a###b")
        self.assertEqual(Parser.remove_whitespaces("a## #b", delimiter="##"), "a## #b")
        self.assertEqual(
            Parser.remove_whitespaces("a## ##b", delimiter="##"), "a## ##b"
        )

    def test_remove_comments(self):
        self.assertEqual(Parser.remove_comments("a#b"), "a")
        self.assertEqual(Parser.remove_comments('"a#b"'), '"a#b"')
        self.assertEqual(Parser.remove_comments('a"#"#b'), 'a"#"')
        self.assertEqual(Parser.remove_comments('a"#""#"#b'), 'a"#""#"')
        self.assertEqual(Parser.remove_comments('#a"#""#"#b'), "")

        self.assertEqual(
            re.sub(
                r"\s",
                "",
                Parser.remove_comments(
                    """
            # a
            b
            # c
            d
        """
                ),
            ),
            "bd",
        )

    async def test_count_brackets(self):
        self.assertEqual(Parser._count_brackets(""), 0)
        self.assertEqual(Parser._count_brackets("()"), 2)
        self.assertEqual(Parser._count_brackets("a()"), 3)
        self.assertEqual(Parser._count_brackets("a(b)"), 4)
        self.assertEqual(Parser._count_brackets("a(b())"), 6)
        self.assertEqual(Parser._count_brackets("a(b(c))"), 7)
        self.assertEqual(Parser._count_brackets("a(b(c))d"), 7)
        self.assertEqual(Parser._count_brackets("a(b(c))d()"), 7)

    def test_split_keyword_arg(self):
        self.assertTupleEqual(Parser._split_keyword_arg("_A=b"), ("_A", "b"))
        self.assertTupleEqual(Parser._split_keyword_arg("a_=1"), ("a_", "1"))
        self.assertTupleEqual(
            Parser._split_keyword_arg("a=repeat(2, KEY_A)"),
            ("a", "repeat(2, KEY_A)"),
        )
        self.assertTupleEqual(Parser._split_keyword_arg('a="=,#+."'), ("a", '"=,#+."'))

    def test_is_this_a_macro(self):
        self.assertTrue(Parser.is_this_a_macro("key(1)"))
        self.assertTrue(Parser.is_this_a_macro("key(1).key(2)"))
        self.assertTrue(Parser.is_this_a_macro("repeat(1, key(1).key(2))"))

        self.assertFalse(Parser.is_this_a_macro("1"))
        self.assertFalse(Parser.is_this_a_macro("key_kp1"))
        self.assertFalse(Parser.is_this_a_macro("btn_left"))
        self.assertFalse(Parser.is_this_a_macro("minus"))
        self.assertFalse(Parser.is_this_a_macro("k"))
        self.assertFalse(Parser.is_this_a_macro(1))
        self.assertFalse(Parser.is_this_a_macro(None))

        self.assertTrue(Parser.is_this_a_macro("a+b"))
        self.assertTrue(Parser.is_this_a_macro("a+b+c"))
        self.assertTrue(Parser.is_this_a_macro("a + b"))
        self.assertTrue(Parser.is_this_a_macro("a + b + c"))

    def test_handle_plus_syntax(self):
        self.assertEqual(Parser.handle_plus_syntax("a + b"), "hold_keys(a,b)")
        self.assertEqual(Parser.handle_plus_syntax("a + b + c"), "hold_keys(a,b,c)")
        self.assertEqual(Parser.handle_plus_syntax(" a+b+c "), "hold_keys(a,b,c)")

        # invalid. The last one with `key` should not have been a parameter
        # of this function to begin with.
        strings = ["+", "a+", "+b", "a\n+\n+\nb", "key(a + b)"]
        for string in strings:
            with self.assertRaises(MacroError):
                logger.info('testing "%s"', string)
                Parser.handle_plus_syntax(string)

        self.assertEqual(Parser.handle_plus_syntax("a"), "a")
        self.assertEqual(Parser.handle_plus_syntax("key(a)"), "key(a)")
        self.assertEqual(Parser.handle_plus_syntax(""), "")

    def test_parse_plus_syntax(self):
        macro = Parser.parse("a + b")
        self.assertEqual(macro.code, "hold_keys(a,b)")

        # this is not erroneously recognized as "plus" syntax
        macro = Parser.parse("key(a) # a + b")
        self.assertEqual(macro.code, "key(a)")

    async def test_extract_params(self):
        # splits strings, doesn't try to understand their meaning yet
        def expect(raw, expectation):
            self.assertListEqual(Parser._extract_args(raw), expectation)

        expect("a", ["a"])
        expect("a,b", ["a", "b"])
        expect("a,b,c", ["a", "b", "c"])

        expect("key(a)", ["key(a)"])
        expect("key(a).key(b), key(a)", ["key(a).key(b)", "key(a)"])
        expect("key(a), key(a).key(b)", ["key(a)", "key(a).key(b)"])

        expect(
            'a("foo(1,2,3)", ",,,,,,    "), , ""',
            ['a("foo(1,2,3)", ",,,,,,    ")', "", '""'],
        )

        expect(
            ",1,   ,b,x(,a(),).y().z(),,",
            ["", "1", "", "b", "x(,a(),).y().z()", "", ""],
        )

        expect("repeat(1, key(a))", ["repeat(1, key(a))"])
        expect(
            "repeat(1, key(a)), repeat(1, key(b))",
            ["repeat(1, key(a))", "repeat(1, key(b))"],
        )
        expect(
            "repeat(1, key(a)), repeat(1, key(b)), repeat(1, key(c))",
            ["repeat(1, key(a))", "repeat(1, key(b))", "repeat(1, key(c))"],
        )

        # will be parsed as None
        expect("", [""])
        expect(",", ["", ""])
        expect(",,", ["", "", ""])

    async def test_parse_params(self):
        def test(value, types):
            argument = Argument(
                ArgumentConfig(position=0, name="test", types=types),
                DummyMapping,
            )
            argument.initialize_variable(RawValue(value=value))
            return argument._variable

        self.assertEqual(
            test("", [None, int, float]),
            Variable(None, const=True),
        )

        # strings. If it is wrapped in quotes, don't parse the contents
        self.assertEqual(
            test('"foo"', [str]),
            Variable("foo", const=True),
        )
        self.assertEqual(
            test('"\tf o o\n"', [str]),
            Variable("\tf o o\n", const=True),
        )
        self.assertEqual(
            test('"foo(a,b)"', [str]),
            Variable("foo(a,b)", const=True),
        )
        self.assertEqual(
            test('",,,()"', [str]),
            Variable(",,,()", const=True),
        )

        # strings without quotes only work as long as there is no function call or
        # anything. This is only really acceptable for constants like KEY_A and for
        # variable names, which are not allowed to contain special characters that may
        # have a meaning in the macro syntax.
        self.assertEqual(
            test("foo", [str]),
            Variable("foo", const=True),
        )

        self.assertEqual(
            test("", [str, None]),
            Variable(None, const=True),
        )
        self.assertEqual(
            test("", [str]),
            Variable("", const=True),
        )
        self.assertEqual(
            test("", [None]),
            Variable(None, const=True),
        )
        self.assertEqual(
            test("None", [None]),
            Variable(None, const=True),
        )
        self.assertEqual(
            test('"None"', [str]),
            Variable("None", const=True),
        )

        self.assertEqual(
            test("5", [int]),
            Variable(5, const=True),
        )
        self.assertEqual(
            test("5", [float, int]),
            Variable(5, const=True),
        )
        self.assertEqual(
            test("5.2", [int, float]),
            Variable(5.2, const=True),
        )
        self.assertIsInstance(
            test("$foo", [str]),
            Variable,
        )
        self.assertEqual(
            test("$foo", [str]),
            Variable("foo", const=False),
        )

    async def test_string_not_a_macro(self):
        # passing a string parameter. This is not a macro, even though
        # it might look like it without the string quotes. Everything with
        # explicit quotes around it has to be treated as a string.
        self.assertRaises(MacroError, Parser.parse, '"modify(a, b)"', self.context)

    async def test_multiline_macro_and_comments(self):
        # the parser is not confused by the code in the comments and can use hashtags
        # in strings in the actual code
        comment = '# repeat(1,key(KEY_D)).set(a,"#b")'
        macro = Parser.parse(
            f"""
            {comment}
            key(KEY_A).{comment}
            key(KEY_B). {comment}
            repeat({comment}
                1, {comment}
                key(KEY_C){comment}
            ). {comment}
            {comment}
            set(a, "#").{comment}
            if_eq($a, "#", key(KEY_E), key(KEY_F)) {comment}
            {comment}
        """,
            self.context,
            DummyMapping,
        )
        await macro.run(self.handler)
        self.assertListEqual(
            self.result,
            [
                (EV_KEY, KEY_A, 1),
                (EV_KEY, KEY_A, 0),
                (EV_KEY, KEY_B, 1),
                (EV_KEY, KEY_B, 0),
                (EV_KEY, KEY_C, 1),
                (EV_KEY, KEY_C, 0),
                (EV_KEY, KEY_E, 1),
                (EV_KEY, KEY_E, 0),
            ],
        )

    async def test_child_macro_count(self):
        # It correctly keeps track of child-macros for both positional and keyword-args
        macro = Parser.parse(
            "hold(macro=hold(hold())).repeat(1, macro=repeat(1, hold()))",
            self.context,
            DummyMapping,
            True,
        )
        self.assertEqual(self.count_child_macros(macro), 4)
        self.assertEqual(self.count_tasks(macro), 6)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import math
from typing import Dict

import evdev
from evdev.ecodes import (
    EV_REL,
    REL_WHEEL,
    REL_HWHEEL,
    REL_WHEEL_HI_RES,
    REL_HWHEEL_HI_RES,
)

from inputremapper import exceptions
from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import (
    Mapping,
    REL_XY_SCALING,
    WHEEL_SCALING,
    WHEEL_HI_RES_SCALING,
)
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.axis_transform import Transformation
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    HandlerEnums,
    InputEventHandler,
)
from inputremapper.input_event import InputEvent
from inputremapper.logging.logger import logger


def is_wheel(event) -> bool:
    return event.type == EV_REL and event.code in (REL_WHEEL, REL_HWHEEL)


def is_high_res_wheel(event) -> bool:
    return event.type == EV_REL and event.code in (REL_WHEEL_HI_RES, REL_HWHEEL_HI_RES)


class Remainder:
    _scale: float
    _remainder: float

    def __init__(self, scale: float):
        self._scale = scale
        self._remainder = 0

    def input(self, value: float) -> int:
        # if the mouse moves very slow, it might not move at all because of the
        # int-conversion (which is required when writing). store the remainder
        # (the decimal places) and add it up, until the mouse moves a little.
        scaled = value * self._scale + self._remainder
        self._remainder = math.fmod(scaled, 1)

        return int(scaled)


class RelToRelHandler(MappingHandler):
    """Handler which transforms EV_REL to EV_REL events."""

    _input_config: InputConfig  # the relative movement we map

    _max_observed_input: float

    _transform: Transformation

    _remainder: Remainder
    _wheel_remainder: Remainder
    _wheel_hi_res_remainder: Remainder

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        global_uinputs: GlobalUInputs,
        **_,
    ) -> None:
        super().__init__(combination, mapping, global_uinputs)

        assert self.mapping.output_code is not None

        # find the input event we are supposed to map. If the input combination is
        # BTN_A + REL_X + BTN_B, then use the value of REL_X for the transformation
        input_config = combination.find_analog_input_config(type_=EV_REL)
        assert input_config is not None
        self._input_config = input_config

        self._max_observed_input = 1

        self._remainder = Remainder(REL_XY_SCALING)
        self._wheel_remainder = Remainder(WHEEL_SCALING)
        self._wheel_hi_res_remainder = Remainder(WHEEL_HI_RES_SCALING)

        self._transform = Transformation(
            max_=1,
            min_=-1,
            deadzone=self.mapping.deadzone,
            gain=self.mapping.gain,
            expo=self.mapping.expo,
        )

    def __str__(self):
        return f"RelToRelHandler for {self._input_config}"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def child(self):  # used for logging
        return f"maps to: {self.mapping.output_code} at {self.mapping.target_uinput}"

    def _should_map(self, event: InputEvent):
        """Check if this input event is relevant for this handler."""
        return event.input_match_hash == self._input_config.input_match_hash

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        if not self._should_map(event):
            return False
        """
        There was the idea to define speed as "movemnt per second".
        There are deprecated mapping variables in this explanation.

        rel2rel example:
        - input every 0.1s (`input_rate` of 10 events/s), value of 200
        - input speed is 2000, because in 1 second a value of 2000 acumulates
        - `input_rel_speed` is a const defined as 4000 px/s, how fast mice usually move
        - `transformed = Transformation(input.value, max=input_rel_speed / input_rate)`
        - get 0.5 because the expo is 0
        - `abs_to_rel_speed` is 5000
        - inject 2500 therefore per second, making it a bit faster
        - divide 2500 by the rate of 10 to inject a value of 250 each time input occurs

        ```
        output_value = Transformation(
            input.value,
            max=input_rel_speed / input_rate
        ) * abs_to_rel_speed / input_rate
        ```

        The input_rel_speed could be used here instead of abs_to_rel_speed, because the
        gain already controls the speed. In that case it would be a 1:1 ratio of
        input-to-output value if the gain is 1.

        for wheel and wheel_hi_res, different input speed constants must be set.

        abs2rel needs a base value for the output, so `abs_to_rel_speed` is still
        required.
        `abs_to_rel_speed / rel_rate * transform(input.value, max=absinfo.max)`
        is the output value. Both abs_to_rel_speed and the transformation-gain control
        speed.

        if abs_to_rel_speed controls speed in the abs2rel output, it should also do so
        in other handlers that have EV_REL output.
        
        unfortunately input_rate needs to be determined during runtime, which screws
        the overall speed up when slowly moving the input device in the beginning,
        because slow input is thought to be the regular input.
        
        ---
        
        transforming from rate based to rate based speed values won't work well.
        
        better to use fractional speed values.
        REL_X of 40 = REL_WHEEL of 1 = REL_WHEE_HI_RES of 1/120
        
        this is why abs_to_rel_speed does not affect the rel_to_rel handler.
        
        The expo calculation will be wrong in the beginning, because it is based on
        the highest observed value. The overall gain will be fine though.
        """

        input_value = float(event.value)

        # scale down now, the remainder calculation scales up by the same factor later
        # depending on what kind of event this becomes.
        if event.is_wheel_event:
            input_value /= WHEEL_SCALING
        elif event.is_wheel_hi_res_event:
            input_value /= WHEEL_HI_RES_SCALING
        else:
            # even though the input rate is unknown we can apply REL_XY_SCALING, which
            # is based on 60hz or something, because the un-scaling also uses values
            # based on 60hz. So the rate cancels out
            input_value /= REL_XY_SCALING

        if abs(input_value) > self._max_observed_input:
            self._max_observed_input = abs(input_value)

        # If _max_observed_input is wrong when the injection starts and the correct
        # value learned during runtime, results can be weird at the beginning.
        # If expo and deadzone are not set, then it is linear and doesn't matter.
        transformed = self._transform(input_value / self._max_observed_input)
        transformed *= self._max_observed_input

        is_wheel_output = self.mapping.is_wheel_output()
        is_hi_res_wheel_output = self.mapping.is_high_res_wheel_output()

        horizontal = self.mapping.output_code in (
            REL_HWHEEL_HI_RES,
            REL_HWHEEL,
        )

        try:
            if is_wheel_output or is_hi_res_wheel_output:
                # inject both kinds of wheels, otherwise wheels don't work for some
                # people. See issue #354
                self._write(
                    REL_HWHEEL if horizontal else REL_WHEEL,
                    self._wheel_remainder.input(transformed),
                )
                self._write(
                    REL_HWHEEL_HI_RES if horizontal else REL_WHEEL_HI_RES,
                    self._wheel_hi_res_remainder.input(transformed),
                )
            else:
                self._write(
                    self.mapping.output_code,
                    self._remainder.input(transformed),
                )

            return True
        except OverflowError:
            # screwed up the calculation of the event value
            logger.error("OverflowError while handling %s", event)
            return True
        except (exceptions.UinputNotAvailable, exceptions.EventNotHandled):
            return False

    def reset(self) -> None:
        pass

    def _write(self, code: int, value: int):
        if value == 0:
            return

        self.global_uinputs.write(
            (EV_REL, code, value),
            self.mapping.target_uinput,
        )

    def needs_wrapping(self) -> bool:
        return len(self.input_configs) > 1

    def set_sub_handler(self, handler: InputEventHandler) -> None:
        assert False  # cannot have a sub-handler

    def wrap_with(self) -> Dict[InputCombination, HandlerEnums]:
        if self.needs_wrapping():
            return {InputCombination(self.input_configs): HandlerEnums.axisswitch}
        return {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import asyncio
import random

from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.task import Task


class WaitTask(Task):
    """Wait time in milliseconds."""

    argument_configs = [
        ArgumentConfig(
            name="time",
            position=0,
            types=[float, int],
        ),
        ArgumentConfig(
            name="max_time",
            position=1,
            types=[float, int, None],
            default=None,
        ),
    ]

    async def run(self, callback) -> None:
        time = self.get_argument("time").get_value()
        max_time = self.get_argument("max_time").get_value()

        if max_time is not None and max_time > time:
            time = random.uniform(time, max_time)

        await asyncio.sleep(time / 1000)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import dataclasses
import functools
import itertools
import unittest
from typing import Iterable, List

from inputremapper.injection.mapping_handlers.axis_transform import Transformation
from tests.lib.test_setup import test_setup


@test_setup
class TestAxisTransformation(unittest.TestCase):
    @dataclasses.dataclass
    class InitArgs:
        max_: int
        min_: int
        deadzone: float
        gain: float
        expo: float

        def values(self):
            return self.__dict__.values()

    def get_init_args(
        self,
        max_=(255, 1000, 2**15),
        min_=(50, 0, -255),
        deadzone=(0, 0.5),
        gain=(0.5, 1, 2),
        expo=(-0.9, 0, 0.3),
    ) -> Iterable[InitArgs]:
        for args in itertools.product(max_, min_, deadzone, gain, expo):
            yield self.InitArgs(*args)

    @staticmethod
    def scale_to_range(min_, max_, x=(-1, -0.2, 0, 0.6, 1)) -> List[float]:
        """Scale values between -1 and 1 up, such that they are between min and max."""
        half_range = (max_ - min_) / 2
        return [float_x * half_range + min_ + half_range for float_x in x]

    def test_scale_to_range(self):
        """Make sure scale_to_range will actually return the min and max values
        (avoid "off by one" errors)"""
        max_ = (255, 1000, 2**15)
        min_ = (50, 0, -255)

        for x1, x2 in itertools.product(min_, max_):
            scaled = self.scale_to_range(x1, x2, (-1, 1))
            self.assertEqual(scaled, [x1, x2])

    def test_expo_symmetry(self):
        """Test that the transformation is symmetric for expo parameter
        x = f(g(x)), if  f._expo == - g._expo

        with the following constraints:
        min = -1, max = 1
        gain = 1
        deadzone = 0

        we can remove the constraints for min, max and gain,
        by scaling the values appropriately after each transformation
        """

        for init_args in self.get_init_args(deadzone=(0,)):
            f = Transformation(*init_args.values())
            init_args.expo = -init_args.expo
            g = Transformation(*init_args.values())

            scale = functools.partial(
                self.scale_to_range,
                init_args.min_,
                init_args.max_,
            )
            for x in scale():
                y1 = g(x)
                y1 = y1 / init_args.gain  # remove the gain
                y1 = scale((y1,))[0]  # remove the min/max constraint

                y2 = f(y1)
                y2 = y2 / init_args.gain  # remove the gain
                y2 = scale((y2,))[0]  # remove the min/max constraint
                self.assertAlmostEqual(x, y2, msg=f"test expo symmetry for {init_args}")

    def test_origin_symmetry(self):
        """Test that the transformation is symmetric to the origin_hash
        f(x) = - f(-x)
        within the constraints: min = -max
        """

        for init_args in self.get_init_args():
            init_args.min_ = -init_args.max_
            f = Transformation(*init_args.values())
            for x in self.scale_to_range(init_args.min_, init_args.max_):
                self.assertAlmostEqual(
                    f(x),
                    -f(-x),
                    msg=f"test origin_hash symmetry at {x=} for {init_args}",
                )

    def test_gain(self):
        """Test that f(max) = gain and f(min) = -gain."""
        for init_args in self.get_init_args():
            f = Transformation(*init_args.values())
            self.assertAlmostEqual(
                f(init_args.max_),
                init_args.gain,
                msg=f"test gain for {init_args}",
            )
            self.assertAlmostEqual(
                f(init_args.min_),
                -init_args.gain,
                msg=f"test gain for {init_args}",
            )

    def test_deadzone(self):
        """Test the Transfomation returns exactly 0 in the range of the deadzone."""

        for init_args in self.get_init_args(deadzone=(0.1, 0.2, 0.9)):
            f = Transformation(*init_args.values())
            for x in self.scale_to_range(
                init_args.min_,
                init_args.max_,
                x=(
                    init_args.deadzone * 0.999,
                    -init_args.deadzone * 0.999,
                    0.3 * init_args.deadzone,
                    0,
                ),
            ):
                self.assertEqual(f(x), 0, msg=f"test deadzone at {x=} for {init_args}")

    def test_continuity_near_deadzone(self):
        """Test that the Transfomation is continues (no sudden jump) next to the
        deadzone"""

        for init_args in self.get_init_args(deadzone=(0.1, 0.2, 0.9)):
            f = Transformation(*init_args.values())
            scale = functools.partial(
                self.scale_to_range,
                init_args.min_,
                init_args.max_,
            )
            x = (
                init_args.deadzone * 1.00001,
                init_args.deadzone * 1.001,
                -init_args.deadzone * 1.00001,
                -init_args.deadzone * 1.001,
            )
            scaled_x = scale(x=x)

            p1 = (x[0], f(scaled_x[0]))  # first point right of deadzone
            p2 = (x[1], f(scaled_x[1]))  # second point right of deadzone

            # calculate a linear function y = m * x + b from p1 and p2
            m = (p1[1] - p2[1]) / (p1[0] - p2[0])
            b = p1[1] - m * p1[0]

            # the zero intersection of that function must be close to the
            # edge of the deadzone
            self.assertAlmostEqual(
                -b / m,
                init_args.deadzone,
                places=5,
                msg=f"test continuity at {init_args.deadzone} for {init_args}",
            )

            # same thing on the other side
            p1 = (x[2], f(scaled_x[2]))
            p2 = (x[3], f(scaled_x[3]))
            m = (p1[1] - p2[1]) / (p1[0] - p2[0])
            b = p1[1] - m * p1[0]
            self.assertAlmostEqual(
                -b / m,
                -init_args.deadzone,
                places=5,
                msg=f"test continuity at {- init_args.deadzone} for {init_args}",
            )

    def test_expo_out_of_range(self):
        f = Transformation(deadzone=0.1, min_=-20, max_=5, expo=1.3)
        self.assertRaises(ValueError, f, 0)
        f = Transformation(deadzone=0.1, min_=-20, max_=5, expo=-1.3)
        self.assertRaises(ValueError, f, 0)

    def test_returns_one_for_range_between_minus_and_plus_one(self):
        for init_args in self.get_init_args(max_=(1,), min_=(-1,), gain=(1,)):
            f = Transformation(*init_args.values())
            self.assertEqual(f(1), 1)
            self.assertEqual(f(-1), -1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import multiprocessing
import os
import select
import time
import unittest

from inputremapper.ipc.pipe import Pipe
from inputremapper.ipc.shared_dict import SharedDict
from inputremapper.ipc.socket import Server, Client, Base
from tests.lib.test_setup import test_setup
from tests.lib.tmp import tmp


@test_setup
class TestSharedDict(unittest.TestCase):
    def setUp(self):
        self.shared_dict = SharedDict()
        self.shared_dict.start()
        time.sleep(0.02)

    def test_returns_none(self):
        self.assertIsNone(self.shared_dict.get("a"))
        self.assertIsNone(self.shared_dict["a"])

    def test_set_get(self):
        self.shared_dict["a"] = 3
        self.assertEqual(self.shared_dict.get("a"), 3)
        self.assertEqual(self.shared_dict["a"], 3)


@test_setup
class TestSocket(unittest.TestCase):
    def test_socket(self):
        def test(s1, s2):
            self.assertEqual(s2.recv(), None)

            s1.send(1)
            self.assertTrue(s2.poll())
            self.assertEqual(s2.recv(), 1)
            self.assertFalse(s2.poll())
            self.assertEqual(s2.recv(), None)

            s1.send(2)
            self.assertTrue(s2.poll())
            s1.send(3)
            self.assertTrue(s2.poll())
            self.assertEqual(s2.recv(), 2)
            self.assertTrue(s2.poll())
            self.assertEqual(s2.recv(), 3)
            self.assertFalse(s2.poll())
            self.assertEqual(s2.recv(), None)

        server = Server(os.path.join(tmp, "socket1"))
        client = Client(os.path.join(tmp, "socket1"))
        test(server, client)

        client = Client(os.path.join(tmp, "socket2"))
        server = Server(os.path.join(tmp, "socket2"))
        test(client, server)

    def test_not_connected_1(self):
        # client discards old message, because it might have had a purpose
        # for a different client and not for the current one
        server = Server(os.path.join(tmp, "socket3"))
        server.send(1)

        client = Client(os.path.join(tmp, "socket3"))
        server.send(2)

        self.assertTrue(client.poll())
        self.assertEqual(client.recv(), 2)
        self.assertFalse(client.poll())
        self.assertEqual(client.recv(), None)

    def test_not_connected_2(self):
        client = Client(os.path.join(tmp, "socket4"))
        client.send(1)

        server = Server(os.path.join(tmp, "socket4"))
        client.send(2)

        self.assertTrue(server.poll())
        self.assertEqual(server.recv(), 2)
        self.assertFalse(server.poll())
        self.assertEqual(server.recv(), None)

    def test_select(self):
        """Is compatible to select.select."""
        server = Server(os.path.join(tmp, "socket6"))
        client = Client(os.path.join(tmp, "socket6"))

        server.send(1)
        ready = select.select([client], [], [], 0)[0][0]
        self.assertEqual(ready, client)

        client.send(2)
        ready = select.select([server], [], [], 0)[0][0]
        self.assertEqual(ready, server)

    def test_base_abstract(self):
        self.assertRaises(NotImplementedError, lambda: Base("foo"))
        self.assertRaises(NotImplementedError, lambda: Base.connect(None))
        self.assertRaises(NotImplementedError, lambda: Base.reconnect(None))
        self.assertRaises(NotImplementedError, lambda: Base.fileno(None))


@test_setup
class TestPipe(unittest.IsolatedAsyncioTestCase):
    def test_pipe_single(self):
        p1 = Pipe(os.path.join(tmp, "pipe"))
        self.assertEqual(p1.recv(), None)

        p1.send(1)
        self.assertTrue(p1.poll())
        self.assertEqual(p1.recv(), 1)
        self.assertFalse(p1.poll())
        self.assertEqual(p1.recv(), None)

        p1.send(2)
        self.assertTrue(p1.poll())
        p1.send(3)
        self.assertTrue(p1.poll())
        self.assertEqual(p1.recv(), 2)
        self.assertTrue(p1.poll())
        self.assertEqual(p1.recv(), 3)
        self.assertFalse(p1.poll())
        self.assertEqual(p1.recv(), None)

    def test_pipe_duo(self):
        p1 = Pipe(os.path.join(tmp, "pipe"))
        p2 = Pipe(os.path.join(tmp, "pipe"))
        self.assertEqual(p2.recv(), None)

        p1.send(1)
        self.assertEqual(p2.recv(), 1)
        self.assertEqual(p2.recv(), None)

        p1.send(2)
        p1.send(3)
        self.assertEqual(p2.recv(), 2)
        self.assertEqual(p2.recv(), 3)
        self.assertEqual(p2.recv(), None)

    async def test_async_for_loop(self):
        p1 = Pipe(os.path.join(tmp, "pipe"))
        iterator = p1.__aiter__()
        p1.send(1)

        self.assertEqual(await iterator.__anext__(), 1)

        read_task = asyncio.Task(iterator.__anext__())
        timeout_task = asyncio.Task(asyncio.sleep(1))

        done, pending = await asyncio.wait(
            (read_task, timeout_task), return_when=asyncio.FIRST_COMPLETED
        )
        self.assertIn(timeout_task, done)
        self.assertIn(read_task, pending)
        read_task.cancel()

    async def test_async_for_loop_duo(self):
        def writer():
            p = Pipe(os.path.join(tmp, "pipe"))
            for i in range(3):
                p.send(i)
            time.sleep(0.5)
            for i in range(3):
                p.send(i)
            time.sleep(0.1)
            p.send("stop now")

        p1 = Pipe(os.path.join(tmp, "pipe"))

        w_process = multiprocessing.Process(target=writer)
        w_process.start()

        messages = []
        async for msg in p1:
            messages.append(msg)
            if msg == "stop now":
                break

        self.assertEqual(messages, [0, 1, 2, 0, 1, 2, "stop now"])


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple, Dict

from inputremapper import exceptions
from inputremapper.configs.input_config import InputCombination
from inputremapper.configs.mapping import Mapping
from inputremapper.exceptions import MappingParsingError
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    HandlerEnums,
)
from inputremapper.input_event import InputEvent
from inputremapper.logging.logger import logger
from inputremapper.utils import get_evdev_constant_name


class KeyHandler(MappingHandler):
    """Injects the target key if notified."""

    _active: bool
    _maps_to: Tuple[int, int]

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        global_uinputs: GlobalUInputs,
        **_,
    ):
        super().__init__(combination, mapping, global_uinputs)
        maps_to = mapping.get_output_type_code()
        if not maps_to:
            raise MappingParsingError(
                "Unable to create key handler from mapping", mapping=mapping
            )

        self._maps_to = maps_to
        self._active = False

    def __str__(self):
        return f"KeyHandler to {self._maps_to}"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def child(self):  # used for logging
        name = get_evdev_constant_name(*self._maps_to)
        return f"maps to: {name} {self._maps_to} on {self.mapping.target_uinput}"

    def notify(self, event: InputEvent, *_, **__) -> bool:
        """Inject event.value to the target key."""
        event_tuple = (*self._maps_to, event.value)
        try:
            self.global_uinputs.write(event_tuple, self.mapping.target_uinput)
            self._active = bool(event.value)
            return True
        except exceptions.Error:
            return False

    def reset(self) -> None:
        logger.debug("resetting key_handler")
        if self._active:
            event_tuple = (*self._maps_to, 0)
            self.global_uinputs.write(event_tuple, self.mapping.target_uinput)
            self._active = False

    def needs_wrapping(self) -> bool:
        return True

    def wrap_with(self) -> Dict[InputCombination, HandlerEnums]:
        return {InputCombination(self.input_configs): HandlerEnums.combination}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from inputremapper.configs.validation_errors import (
    MacroError,
)
from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.macro import macro_variables
from inputremapper.injection.macros.parse import Parser
from inputremapper.injection.macros.raw_value import RawValue
from inputremapper.injection.macros.task import Task
from tests.unit.test_macros.macro_test_base import DummyMapping, MacroTestBase


class TestDynamicTypes(MacroTestBase):
    # "Dynamic" meaning const=False
    async def test_set_type_int(self):
        await Parser.parse(
            "set(a, 1)",
            self.context,
            DummyMapping,
            True,
        ).run(lambda *_, **__: None)
        self.assertEqual(macro_variables.get("a"), 1)
        # assertEqual(1.0, 1) passes, so check for the type to be sure:
        self.assertIsInstance(macro_variables.get("a"), int)

    async def test_set_type_float(self):
        await Parser.parse(
            "set(a, 2.2)",
            self.context,
            DummyMapping,
            True,
        ).run(lambda *_, **__: None)
        self.assertEqual(macro_variables.get("a"), 2.2)

    async def test_set_type_str(self):
        await Parser.parse(
            'set(a, "3")',
            self.context,
            DummyMapping,
            True,
        ).run(lambda *_, **__: None)
        self.assertEqual(macro_variables.get("a"), "3")

    def make_test_task(self, types):
        # Make a new test task, with a different types array each time.
        class TestTask(Task):
            argument_configs = [
                ArgumentConfig(
                    name="testvalue",
                    position=0,
                    types=types,
                )
            ]

        return TestTask(
            [RawValue("$a")],
            {},
            self.context,
            DummyMapping,
        )

    async def test_dynamic_int_parsing(self):
        # set(a, 4) was used. Could be meant as an integer, or as a string
        # (just like how key(KEY_A) doesn't require string quotes to be a string)
        macro_variables["a"] = 4

        test_task = self.make_test_task([str, int])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), 4)

        test_task = self.make_test_task([int])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), 4)

        # Now that ints are not allowed, it will be used as a string
        test_task = self.make_test_task([str])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), "4")

    async def test_dynamic_float_parsing(self):
        # set(a, 5.5) was used.
        macro_variables["a"] = 5.5

        test_task = self.make_test_task([str, float])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), 5.5)

        test_task = self.make_test_task([float])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), 5.5)

        test_task = self.make_test_task([str])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), "5.5")

    async def test_no_float_allowed(self):
        # set(a, 6.6) was used.
        macro_variables["a"] = 6.6

        test_task = self.make_test_task([str, int])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), "6.6")

        test_task = self.make_test_task([int])
        self.assertRaises(
            MacroError,
            lambda: test_task.get_argument("testvalue").get_value(),
        )

    async def test_force_string_float(self):
        # set(a, "7.7") was used. Since quotes are explicitly added, the variable is
        # not intended to be used as a float.
        macro_variables["a"] = "7.7"

        test_task = self.make_test_task([str, float])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), "7.7")

        test_task = self.make_test_task([float])
        self.assertRaises(
            MacroError,
            lambda: test_task.get_argument("testvalue").get_value(),
        )

    async def test_force_string_int(self):
        # set(a, "8") was used.
        macro_variables["a"] = "8"

        test_task = self.make_test_task([int, str])
        self.assertEqual(test_task.get_argument("testvalue").get_value(), "8")

        test_task = self.make_test_task([int])
        self.assertRaises(
            MacroError,
            lambda: test_task.get_argument("testvalue").get_value(),
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import os.path
import unittest
from typing import List
from unittest.mock import patch, MagicMock, call

import gi
from evdev.ecodes import EV_ABS, ABS_X, ABS_Y, ABS_RX

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.injection.injector import InjectorState

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.groups import _Groups
from inputremapper.gui.messages.message_broker import (
    MessageBroker,
    MessageType,
)
from inputremapper.gui.messages.message_data import (
    GroupsData,
    GroupData,
    PresetData,
    StatusData,
    CombinationRecorded,
    CombinationUpdate,
    UserConfirmRequest,
)
from inputremapper.gui.reader_client import ReaderClient
from inputremapper.gui.utils import CTX_ERROR, CTX_APPLY, gtk_iteration
from inputremapper.gui.gettext import _
from inputremapper.injection.global_uinputs import GlobalUInputs, FrontendUInput
from inputremapper.configs.mapping import UIMapping, MappingData, Mapping
from tests.lib.spy import spy
from tests.lib.patches import FakeDaemonProxy
from tests.lib.fixtures import fixtures, prepare_presets
from inputremapper.configs.global_config import GlobalConfig
from inputremapper.gui.controller import Controller, MAPPING_DEFAULTS
from inputremapper.gui.data_manager import DataManager, DEFAULT_PRESET_NAME
from inputremapper.configs.paths import PathUtils
from inputremapper.configs.preset import Preset
from tests.lib.test_setup import test_setup


@test_setup
class TestController(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.message_broker = MessageBroker()
        uinputs = GlobalUInputs(FrontendUInput)
        uinputs.prepare_all()
        self.data_manager = DataManager(
            self.message_broker,
            GlobalConfig(),
            ReaderClient(self.message_broker, _Groups()),
            FakeDaemonProxy(),
            uinputs,
            keyboard_layout,
        )
        self.user_interface = MagicMock()
        self.controller = Controller(self.message_broker, self.data_manager)
        self.controller.set_gui(self.user_interface)

    def test_should_get_newest_group(self):
        """get_a_group should the newest group."""
        with patch.object(
            self.data_manager, "get_newest_group_key", MagicMock(return_value="foo")
        ):
            self.assertEqual(self.controller.get_a_group(), "foo")

    def test_should_get_any_group(self):
        """get_a_group should return a valid group."""
        with patch.object(
            self.data_manager,
            "get_newest_group_key",
            MagicMock(side_effect=FileNotFoundError),
        ):
            fixture_keys = [fixture.group_key or fixture.name for fixture in fixtures]
            self.assertIn(self.controller.get_a_group(), fixture_keys)

    def test_should_get_newest_preset(self):
        """get_a_group should the newest group."""
        with patch.object(
            self.data_manager, "get_newest_preset_name", MagicMock(return_value="bar")
        ):
            self.data_manager.load_group("Foo Device")
            self.assertEqual(self.controller.get_a_preset(), "bar")

    def test_should_get_any_preset(self):
        """get_a_preset should return a new preset if none exist."""
        self.data_manager.load_group("Foo Device")
        # the default name
        self.assertEqual(self.controller.get_a_preset(), "new preset")

    def test_on_init_should_provide_uinputs(self):
        calls = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.uinputs, f)
        self.message_broker.signal(MessageType.init)
        self.assertEqual(
            ["keyboard", "gamepad", "mouse", "keyboard + mouse"],
            list(calls[-1].uinputs.keys()),
        )

    def test_on_init_should_provide_groups(self):
        calls: List[GroupsData] = []

        def f(groups):
            calls.append(groups)

        self.message_broker.subscribe(MessageType.groups, f)
        self.message_broker.signal(MessageType.init)
        self.assertEqual(
            ["Foo Device", "Foo Device 2", "Bar Device", "gamepad", "Qux/[Device]?"],
            list(calls[-1].groups.keys()),
        )

    def test_on_init_should_provide_a_group(self):
        calls: List[GroupData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.group, f)
        self.message_broker.signal(MessageType.init)
        self.assertGreaterEqual(len(calls), 1)

    def test_on_init_should_provide_a_preset(self):
        calls: List[PresetData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.preset, f)
        self.message_broker.signal(MessageType.init)
        self.assertGreaterEqual(len(calls), 1)

    def test_on_init_should_provide_a_mapping(self):
        """Only if there is one."""
        prepare_presets()
        calls: List[MappingData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.mapping, f)
        self.message_broker.signal(MessageType.init)
        self.assertTrue(calls[-1].is_valid())

    def test_on_init_should_provide_a_default_mapping(self):
        """If there is no real preset available"""
        calls: List[MappingData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.mapping, f)
        self.message_broker.signal(MessageType.init)
        for m in calls:
            self.assertEqual(m, UIMapping(**MAPPING_DEFAULTS))

    def test_on_load_group_should_provide_preset(self):
        with patch.object(self.data_manager, "load_preset") as mock:
            self.controller.load_group("Foo Device")
            mock.assert_called_once()

    def test_on_load_group_should_provide_mapping(self):
        """If there is one"""
        prepare_presets()
        calls: List[MappingData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.mapping, f)
        self.controller.load_group(group_key="Foo Device 2")
        self.assertTrue(calls[-1].is_valid())

    def test_on_load_group_should_provide_default_mapping(self):
        """If there is none."""
        calls: List[MappingData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.mapping, f)

        self.controller.load_group(group_key="Foo Device")
        for m in calls:
            self.assertEqual(m, UIMapping(**MAPPING_DEFAULTS))

    def test_on_load_preset_should_provide_mapping(self):
        """If there is one."""
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        calls: List[MappingData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.mapping, f)
        self.controller.load_preset(name="preset2")
        self.assertTrue(calls[-1].is_valid())

    def test_on_load_preset_should_provide_default_mapping(self):
        """If there is none."""
        Preset(PathUtils.get_preset_path("Foo Device", "bar")).save()
        self.data_manager.load_group("Foo Device 2")
        calls: List[MappingData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.mapping, f)
        self.controller.load_preset(name="bar")
        for m in calls:
            self.assertEqual(m, UIMapping(**MAPPING_DEFAULTS))

    def test_on_delete_preset_asks_for_confirmation(self):
        prepare_presets()
        self.message_broker.signal(MessageType.init)
        mock = MagicMock()
        self.message_broker.subscribe(MessageType.user_confirm_request, mock)
        self.controller.delete_preset()
        mock.assert_called_once()

    def test_deletes_preset_when_confirmed(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.message_broker.subscribe(
            MessageType.user_confirm_request, lambda msg: msg.respond(True)
        )
        self.controller.delete_preset()
        self.assertFalse(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

    def test_does_not_delete_preset_when_not_confirmed(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.user_interface.confirm_delete.configure_mock(
            return_value=Gtk.ResponseType.CANCEL
        )
        self.controller.delete_preset()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

    def test_copy_preset(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()

        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy"))
        )

    def test_copy_preset_should_add_number(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()  # creates "preset2 copy"
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()  # creates "preset2 copy 2"

        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy 2"))
        )

    def test_copy_preset_should_increment_existing_number(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()  # creates "preset2 copy"
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()  # creates "preset2 copy 2"
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()  # creates "preset2 copy 3"

        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy 2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy 3"))
        )

    def test_copy_preset_should_not_append_copy_twice(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()  # creates "preset2 copy"
        self.controller.copy_preset()  # creates "preset2 copy 2" not "preset2 copy copy"

        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy 2"))
        )

    def test_copy_preset_should_not_append_copy_to_copy_with_number(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()  # creates "preset2 copy"
        self.data_manager.load_preset("preset2")
        self.controller.copy_preset()  # creates "preset2 copy 2"
        self.controller.copy_preset()  # creates "preset2 copy 3" not "preset2 copy 2 copy"

        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy 2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 copy 3"))
        )

    def test_rename_preset(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertFalse(os.path.exists(PathUtils.get_preset_path("Foo Device", "foo")))

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.controller.rename_preset(new_name="foo")

        self.assertFalse(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertTrue(os.path.exists(PathUtils.get_preset_path("Foo Device", "foo")))

    def test_rename_preset_sanitized(self):
        Preset(PathUtils.get_preset_path("Qux/[Device]?", "bla")).save()

        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Qux/[Device]?", "bla"))
        )
        self.assertFalse(
            os.path.exists(PathUtils.get_preset_path("Qux/[Device]?", "blubb"))
        )

        self.data_manager.load_group("Qux/[Device]?")
        self.data_manager.load_preset("bla")
        self.controller.rename_preset(new_name="foo:/bar")

        # all functions expect the true name, which is also shown to the user, but on
        # the file system it always uses sanitized names.
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Qux/[Device]?", "foo__bar"))
        )

        # since the name is never stored in an un-sanitized way, this can't work
        self.assertFalse(
            os.path.exists(PathUtils.get_preset_path("Qux/[Device]?", "foo:/bar"))
        )

        path = os.path.join(
            PathUtils.config_path(), "presets", "Qux_[Device]_", "foo__bar.json"
        )
        self.assertTrue(os.path.exists(path))

        # using the sanitized name in function calls works as well
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Qux_[Device]_", "foo__bar"))
        )

    def test_rename_preset_should_pick_available_name(self):
        prepare_presets()
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset3"))
        )
        self.assertFalse(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset3 2"))
        )

        self.data_manager.load_group("Foo Device")
        self.data_manager.load_preset("preset2")
        self.controller.rename_preset(new_name="preset3")

        self.assertFalse(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset3"))
        )
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset3 2"))
        )

    def test_rename_preset_should_not_rename_to_empty_name(self):
        prepare_presets()
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device")
        self.data_manager.load_preset("preset2")
        self.controller.rename_preset(new_name="")

        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

    def test_rename_preset_should_not_update_same_name(self):
        """When the new name is the same as the current name."""
        prepare_presets()
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.controller.rename_preset(new_name="preset2")

        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2"))
        )
        self.assertFalse(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "preset2 2"))
        )

    def test_on_add_preset_uses_default_name(self):
        self.assertFalse(
            os.path.exists(PathUtils.get_preset_path("Foo Device", DEFAULT_PRESET_NAME))
        )

        self.data_manager.load_group("Foo Device 2")

        self.controller.add_preset()
        self.assertTrue(
            os.path.exists(PathUtils.get_preset_path("Foo Device", "new preset"))
        )

    def test_on_add_preset_uses_provided_name(self):
        self.assertFalse(os.path.exists(PathUtils.get_preset_path("Foo Device", "foo")))

        self.data_manager.load_group("Foo Device 2")

        self.controller.add_preset(name="foo")
        self.assertTrue(os.path.exists(PathUtils.get_preset_path("Foo Device", "foo")))

    def test_on_add_preset_shows_permission_error_status(self):
        self.data_manager.load_group("Foo Device 2")

        msg = None

        def f(data):
            nonlocal msg
            msg = data

        self.message_broker.subscribe(MessageType.status_msg, f)
        mock = MagicMock(side_effect=PermissionError)
        with patch("inputremapper.configs.preset.Preset.save", mock):
            self.controller.add_preset("foo")

        mock.assert_called()
        self.assertIsNotNone(msg)
        self.assertIn("Permission denied", msg.msg)

    def test_on_update_mapping(self):
        """Update_mapping should call data_manager.update_mapping.

        This ensures mapping_changed is emitted.
        """
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            combination=InputCombination([InputConfig(type=1, code=4)])
        )

        with patch.object(self.data_manager, "update_mapping") as mock:
            self.controller.update_mapping(
                name="foo",
                output_symbol="f",
                release_timeout=0.3,
            )
            mock.assert_called_once()

    def test_create_mapping_will_load_the_created_mapping(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")

        calls: List[MappingData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.mapping, f)
        self.controller.create_mapping()

        self.assertEqual(calls[-1], UIMapping(**MAPPING_DEFAULTS))

    def test_create_mapping_should_not_create_multiple_empty_mappings(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.controller.create_mapping()  # create a first empty mapping

        calls = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.mapping, f)
        self.message_broker.subscribe(MessageType.preset, f)

        self.controller.create_mapping()  # try to create a second one
        self.assertEqual(len(calls), 0)

    def test_delete_mapping_asks_for_confirmation(self):
        prepare_presets()
        self.message_broker.signal(MessageType.init)
        mock = MagicMock()
        self.message_broker.subscribe(MessageType.user_confirm_request, mock)
        self.controller.delete_mapping()
        mock.assert_called_once()

    def test_deletes_mapping_when_confirmed(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(InputCombination([InputConfig(type=1, code=3)]))
        self.message_broker.subscribe(
            MessageType.user_confirm_request, lambda msg: msg.respond(True)
        )
        self.controller.delete_mapping()
        self.controller.save()

        preset = Preset(PathUtils.get_preset_path("Foo Device", "preset2"))
        preset.load()
        self.assertIsNone(
            preset.get_mapping(InputCombination([InputConfig(type=1, code=3)]))
        )

    def test_does_not_delete_mapping_when_not_confirmed(self):
        prepare_presets()
        self.assertTrue(
            os.path.isfile(PathUtils.get_preset_path("Foo Device", "preset2"))
        )

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(InputCombination([InputConfig(type=1, code=3)]))
        self.user_interface.confirm_delete.configure_mock(
            return_value=Gtk.ResponseType.CANCEL
        )

        self.controller.delete_mapping()
        self.controller.save()

        preset = Preset(PathUtils.get_preset_path("Foo Device", "preset2"))
        preset.load()
        self.assertIsNotNone(
            preset.get_mapping(InputCombination([InputConfig(type=1, code=3)]))
        )

    def test_should_update_combination(self):
        """When combination is free."""
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(InputCombination([InputConfig(type=1, code=3)]))

        calls: List[CombinationUpdate] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.combination_update, f)
        self.controller.update_combination(
            InputCombination([InputConfig(type=1, code=10)])
        )
        self.assertEqual(
            calls[0],
            CombinationUpdate(
                InputCombination([InputConfig(type=1, code=3)]),
                InputCombination([InputConfig(type=1, code=10)]),
            ),
        )

    def test_should_not_update_combination(self):
        """When combination is already used."""
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(InputCombination([InputConfig(type=1, code=3)]))

        calls: List[CombinationUpdate] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.combination_update, f)
        self.controller.update_combination(
            InputCombination([InputConfig(type=1, code=4)])
        )
        self.assertEqual(len(calls), 0)

    def test_sets_input_to_analog(self):
        prepare_presets()

        input_config = InputConfig(type=EV_ABS, code=ABS_RX)

        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.active_preset.add(
            Mapping(
                input_combination=InputCombination([input_config]),
                output_type=EV_ABS,
                output_code=ABS_X,
                target_uinput="gamepad",
            )
        )
        self.data_manager.load_mapping(InputCombination([input_config]))

        self.controller.start_key_recording()
        self.message_broker.publish(
            CombinationRecorded(
                InputCombination(
                    [
                        InputConfig(
                            type=EV_ABS,
                            code=ABS_Y,
                            analog_threshold=50,
                        ),
                        InputConfig(
                            type=EV_ABS,
                            code=ABS_RX,
                            analog_threshold=60,
                        ),
                    ]
                )
            )
        )

        # the analog_threshold is removed automatically, otherwise the mapping doesn't
        # make sense because only analog inputs can map to analog outputs.
        # This is indicated by is_analog_output being true.
        self.assertTrue(self.controller.data_manager.active_mapping.is_analog_output())

        # only the first input is modified
        active_mapping = self.controller.data_manager.active_mapping
        self.assertEqual(active_mapping.input_combination[0].analog_threshold, None)
        self.assertEqual(active_mapping.input_combination[1].analog_threshold, 60)

    def test_key_recording_disables_gui_shortcuts(self):
        self.message_broker.signal(MessageType.init)
        self.user_interface.disconnect_shortcuts.assert_not_called()
        self.controller.start_key_recording()
        self.user_interface.disconnect_shortcuts.assert_called_once()

    def test_key_recording_enables_gui_shortcuts_when_finished(self):
        self.message_broker.signal(MessageType.init)
        self.controller.start_key_recording()

        self.user_interface.connect_shortcuts.assert_not_called()
        self.message_broker.signal(MessageType.recording_finished)
        self.user_interface.connect_shortcuts.assert_called_once()

    def test_key_recording_enables_gui_shortcuts_when_stopped(self):
        self.message_broker.signal(MessageType.init)
        self.controller.start_key_recording()

        self.user_interface.connect_shortcuts.assert_not_called()
        self.controller.stop_key_recording()
        self.user_interface.connect_shortcuts.assert_called_once()

    def test_recording_messages(self):
        mock1 = MagicMock()
        mock2 = MagicMock()
        self.message_broker.subscribe(MessageType.recording_started, mock1)
        self.message_broker.subscribe(MessageType.recording_finished, mock2)

        self.message_broker.signal(MessageType.init)
        self.controller.start_key_recording()

        mock1.assert_called_once()
        mock2.assert_not_called()

        self.controller.stop_key_recording()

        mock1.assert_called_once()
        mock2.assert_called_once()

    def test_key_recording_updates_mapping_combination(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(InputCombination([InputConfig(type=1, code=3)]))

        calls: List[CombinationUpdate] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.combination_update, f)

        self.controller.start_key_recording()
        self.message_broker.publish(
            CombinationRecorded(InputCombination([InputConfig(type=1, code=10)]))
        )
        self.assertEqual(
            calls[0],
            CombinationUpdate(
                InputCombination([InputConfig(type=1, code=3)]),
                InputCombination([InputConfig(type=1, code=10)]),
            ),
        )
        self.message_broker.publish(
            CombinationRecorded(
                InputCombination(InputCombination.from_tuples((1, 10), (1, 3)))
            )
        )
        self.assertEqual(
            calls[1],
            CombinationUpdate(
                InputCombination([InputConfig(type=1, code=10)]),
                InputCombination(InputCombination.from_tuples((1, 10), (1, 3))),
            ),
        )

    def test_no_key_recording_when_not_started(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(InputCombination([InputConfig(type=1, code=3)]))

        calls: List[CombinationUpdate] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.combination_update, f)

        self.message_broker.publish(
            CombinationRecorded(InputCombination([InputConfig(type=1, code=10)]))
        )
        self.assertEqual(len(calls), 0)

    def test_key_recording_stops_when_finished(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(InputCombination([InputConfig(type=1, code=3)]))

        calls: List[CombinationUpdate] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.combination_update, f)

        self.controller.start_key_recording()
        self.message_broker.publish(
            CombinationRecorded(InputCombination([InputConfig(type=1, code=10)]))
        )
        self.message_broker.signal(MessageType.recording_finished)
        self.message_broker.publish(
            CombinationRecorded(
                InputCombination(InputCombination.from_tuples((1, 10), (1, 3)))
            )
        )

        self.assertEqual(len(calls), 1)  # only the first was processed

    def test_key_recording_stops_when_stopped(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(InputCombination([InputConfig(type=1, code=3)]))

        calls: List[CombinationUpdate] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.combination_update, f)

        self.controller.start_key_recording()
        self.message_broker.publish(
            CombinationRecorded(InputCombination([InputConfig(type=1, code=10)]))
        )
        self.controller.stop_key_recording()
        self.message_broker.publish(
            CombinationRecorded(
                InputCombination(InputCombination.from_tuples((1, 10), (1, 3)))
            )
        )

        self.assertEqual(len(calls), 1)  # only the first was processed

    def test_start_injecting_shows_status_when_preset_empty(self):
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.create_preset("foo")
        self.data_manager.load_preset("foo")
        calls: List[StatusData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.status_msg, f)

        def f2():
            raise AssertionError("Injection started unexpectedly")

        self.data_manager.start_injecting = f2
        self.controller.start_injecting()

        self.assertEqual(
            calls[-1], StatusData(CTX_ERROR, _("You need to add mappings first"))
        )

    def test_start_injecting_warns_about_btn_left(self):
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.create_preset("foo")
        self.data_manager.load_preset("foo")
        self.data_manager.create_mapping()
        self.data_manager.update_mapping(
            input_combination=InputCombination([InputConfig.btn_left()]),
            target_uinput="keyboard",
            output_symbol="a",
        )
        calls: List[StatusData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.status_msg, f)

        def f2():
            raise AssertionError("Injection started unexpectedly")

        self.data_manager.start_injecting = f2
        self.controller.start_injecting()

        self.assertEqual(calls[-1].ctx_id, CTX_ERROR)
        self.assertIn("BTN_LEFT", calls[-1].tooltip)

    def test_start_injecting_starts_with_btn_left_on_second_try(self):
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.create_preset("foo")
        self.data_manager.load_preset("foo")
        self.data_manager.create_mapping()
        self.data_manager.update_mapping(
            input_combination=InputCombination([InputConfig.btn_left()]),
            target_uinput="keyboard",
            output_symbol="a",
        )

        with patch.object(self.data_manager, "start_injecting") as mock:
            self.controller.start_injecting()
            mock.assert_not_called()
            self.controller.start_injecting()
            mock.assert_called_once()

    def test_start_injecting_starts_with_btn_left_when_mapped_to_other_button(self):
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.create_preset("foo")
        self.data_manager.load_preset("foo")
        self.data_manager.create_mapping()
        self.data_manager.update_mapping(
            input_combination=InputCombination([InputConfig.btn_left()]),
            target_uinput="keyboard",
            output_symbol="a",
        )
        self.data_manager.create_mapping()
        self.data_manager.load_mapping(InputCombination.empty_combination())
        self.data_manager.update_mapping(
            input_combination=InputCombination([InputConfig(type=1, code=5)]),
            target_uinput="mouse",
            output_symbol="BTN_LEFT",
        )

        mock = MagicMock(return_value=True)
        self.data_manager.start_injecting = mock
        self.controller.start_injecting()
        mock.assert_called()

    def test_start_injecting_shows_status(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        calls: List[StatusData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.status_msg, f)
        mock = MagicMock(return_value=True)
        self.data_manager.start_injecting = mock
        self.controller.start_injecting()

        mock.assert_called()
        self.assertEqual(calls[0], StatusData(CTX_APPLY, _("Starting injection...")))

    def test_start_injecting_shows_failure_status(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        calls: List[StatusData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.status_msg, f)
        mock = MagicMock(return_value=False)
        self.data_manager.start_injecting = mock
        self.controller.start_injecting()

        mock.assert_called()
        self.assertEqual(
            calls[-1],
            StatusData(
                CTX_APPLY,
                _('Failed to apply preset "%s"') % self.data_manager.active_preset.name,
            ),
        )

    def test_start_injecting_adds_listener_to_update_injector_status(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")

        with patch.object(self.message_broker, "subscribe") as mock:
            self.controller.start_injecting()
            mock.assert_called_once_with(
                MessageType.injector_state, self.controller.show_injector_result
            )

    def test_stop_injecting_shows_status(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        calls: List[StatusData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.status_msg, f)
        mock = MagicMock(return_value=InjectorState.STOPPED)
        self.data_manager.get_state = mock
        self.controller.stop_injecting()
        gtk_iteration(50)

        mock.assert_called()
        self.assertEqual(calls[-1], StatusData(CTX_APPLY, _("Stopped the injection")))

    def test_show_injection_result(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")

        mock = MagicMock(return_value=InjectorState.RUNNING)
        self.data_manager.get_state = mock
        calls: List[StatusData] = []

        def f(data):
            calls.append(data)

        self.message_broker.subscribe(MessageType.status_msg, f)

        self.controller.start_injecting()
        gtk_iteration(50)
        self.assertEqual(calls[-1].msg, _('Applied preset "%s"') % "preset2")

        mock.return_value = InjectorState.ERROR
        self.controller.start_injecting()
        gtk_iteration(50)
        self.assertEqual(calls[-1].msg, _('Error applying preset "%s"') % "preset2")

        mock.return_value = InjectorState.NO_GRAB
        self.controller.start_injecting()
        gtk_iteration(50)
        self.assertEqual(calls[-1].msg, _('Failed to apply preset "%s"') % "preset2")

        mock.return_value = InjectorState.UPGRADE_EVDEV
        self.controller.start_injecting()
        gtk_iteration(50)
        self.assertEqual(calls[-1].msg, "Upgrade python-evdev")

    def test_close(self):
        mock_save = MagicMock()
        listener = MagicMock()
        self.message_broker.subscribe(MessageType.terminate, listener)
        self.data_manager.save = mock_save

        self.controller.close()
        mock_save.assert_called()
        listener.assert_called()

    def test_set_autoload_refreshes_service_config(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")

        with patch.object(self.data_manager, "refresh_service_config_path") as mock:
            self.controller.set_autoload(True)
            mock.assert_called()

    def test_move_event_up(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination(
                InputCombination.from_tuples((1, 1), (1, 2), (1, 3))
            )
        )

        self.controller.move_input_config_in_combination(
            InputConfig(type=1, code=2), "up"
        )
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((1, 2), (1, 1), (1, 3))),
        )
        # now nothing changes
        self.controller.move_input_config_in_combination(
            InputConfig(type=1, code=2), "up"
        )
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((1, 2), (1, 1), (1, 3))),
        )

    def test_move_event_down(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination(
                InputCombination.from_tuples((1, 1), (1, 2), (1, 3))
            )
        )

        self.controller.move_input_config_in_combination(
            InputConfig(type=1, code=2), "down"
        )
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((1, 1), (1, 3), (1, 2))),
        )
        # now nothing changes
        self.controller.move_input_config_in_combination(
            InputConfig(type=1, code=2), "down"
        )
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((1, 1), (1, 3), (1, 2))),
        )

    def test_move_event_in_combination_of_len_1(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.controller.move_input_config_in_combination(
            InputConfig(type=1, code=3), "down"
        )
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((1, 3))),
        )

    def test_move_event_loads_it_again(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination(
                InputCombination.from_tuples((1, 1), (1, 2), (1, 3))
            )
        )
        mock = MagicMock()
        self.message_broker.subscribe(MessageType.selected_event, mock)
        self.controller.move_input_config_in_combination(
            InputConfig(type=1, code=2), "down"
        )
        mock.assert_called_once_with(InputConfig(type=1, code=2))

    def test_update_event(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.load_input_config(InputConfig(type=1, code=3))
        mock = MagicMock()
        self.message_broker.subscribe(MessageType.selected_event, mock)
        self.controller.update_input_config(InputConfig(type=1, code=10))
        mock.assert_called_once_with(InputConfig(type=1, code=10))

    def test_update_event_reloads_mapping_and_event_when_update_fails(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.load_input_config(InputConfig(type=1, code=3))
        mock = MagicMock()
        self.message_broker.subscribe(MessageType.selected_event, mock)
        self.message_broker.subscribe(MessageType.mapping, mock)
        calls = [
            call(self.data_manager.active_mapping.get_bus_message()),
            call(InputConfig(type=1, code=3)),
        ]
        self.controller.update_input_config(
            InputConfig(type=1, code=4)
        )  # already exists
        mock.assert_has_calls(calls, any_order=False)

    def test_remove_event_does_nothing_when_mapping_not_loaded(self):
        with spy(self.data_manager, "update_mapping") as mock:
            self.controller.remove_event()
            mock.assert_not_called()

    def test_remove_event_removes_active_event(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (1, 4))
        )
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((1, 3), (1, 4))),
        )
        self.data_manager.load_input_config(InputConfig(type=1, code=4))

        self.controller.remove_event()
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((1, 3))),
        )

    def test_remove_event_loads_a_event(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (1, 4))
        )
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((1, 3), (1, 4))),
        )
        self.data_manager.load_input_config(InputConfig(type=1, code=4))

        mock = MagicMock()
        self.message_broker.subscribe(MessageType.selected_event, mock)
        self.controller.remove_event()
        mock.assert_called_once_with(InputConfig(type=1, code=3))

    def test_remove_event_reloads_mapping_and_event_when_update_fails(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (1, 4))
        )
        self.data_manager.load_input_config(InputConfig(type=1, code=3))

        # removing "1,3,1" will throw a key error because a mapping with combination
        # "1,4,1" already exists in preset
        mock = MagicMock()
        self.message_broker.subscribe(MessageType.selected_event, mock)
        self.message_broker.subscribe(MessageType.mapping, mock)
        calls = [
            call(self.data_manager.active_mapping.get_bus_message()),
            call(InputConfig(type=1, code=3)),
        ]
        self.controller.remove_event()
        mock.assert_has_calls(calls, any_order=False)

    def test_set_event_as_analog_saves(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((3, 0, 10))
        )
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((3, 0, 10)))
        )
        self.data_manager.load_input_config(
            InputConfig(type=3, code=0, analog_threshold=10)
        )

        with patch.object(self.data_manager, "save") as mock:
            self.controller.set_event_as_analog(False)
            mock.assert_called_once()

        with patch.object(self.data_manager, "save") as mock:
            self.controller.set_event_as_analog(True)
            mock.assert_called_once()

    def test_set_event_as_analog_sets_input_to_analog(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((3, 0, 10))
        )
        self.data_manager.load_input_config(
            InputConfig(type=3, code=0, analog_threshold=10)
        )

        self.controller.set_event_as_analog(True)
        self.assertEqual(
            self.data_manager.active_mapping.input_combination,
            InputCombination(InputCombination.from_tuples((3, 0))),
        )

    def test_set_event_as_analog_adds_rel_threshold(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((2, 0))
        )
        self.data_manager.load_input_config(InputConfig(type=2, code=0))

        self.controller.set_event_as_analog(False)
        combinations = [
            InputCombination(InputCombination.from_tuples((2, 0, 1))),
            InputCombination(InputCombination.from_tuples((2, 0, -1))),
        ]
        self.assertIn(self.data_manager.active_mapping.input_combination, combinations)

    def test_set_event_as_analog_adds_abs_threshold(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((3, 0))
        )
        self.data_manager.load_input_config(InputConfig(type=3, code=0))

        self.controller.set_event_as_analog(False)
        combinations = [
            InputCombination(InputCombination.from_tuples((3, 0, 10))),
            InputCombination(InputCombination.from_tuples((3, 0, -10))),
        ]
        self.assertIn(self.data_manager.active_mapping.input_combination, combinations)

    def test_set_event_as_analog_reloads_mapping_and_event_when_key_event(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.load_input_config(InputConfig(type=1, code=3))

        mock = MagicMock()
        self.message_broker.subscribe(MessageType.selected_event, mock)
        self.message_broker.subscribe(MessageType.mapping, mock)
        calls = [
            call(self.data_manager.active_mapping.get_bus_message()),
            call(InputConfig(type=1, code=3)),
        ]
        self.controller.set_event_as_analog(True)
        mock.assert_has_calls(calls, any_order=False)

    def test_set_event_as_analog_reloads_when_setting_to_analog_fails(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((3, 0, 10))
        )
        self.data_manager.load_input_config(
            InputConfig(type=3, code=0, analog_threshold=10)
        )

        mock = MagicMock()
        self.message_broker.subscribe(MessageType.selected_event, mock)
        self.message_broker.subscribe(MessageType.mapping, mock)
        calls = [
            call(self.data_manager.active_mapping.get_bus_message()),
            call(InputConfig(type=3, code=0, analog_threshold=10)),
        ]
        with patch.object(self.data_manager, "update_mapping", side_effect=KeyError):
            self.controller.set_event_as_analog(True)
            mock.assert_has_calls(calls, any_order=False)

    def test_set_event_as_analog_reloads_when_setting_to_key_fails(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((3, 0))
        )
        self.data_manager.load_input_config(InputConfig(type=3, code=0))

        mock = MagicMock()
        self.message_broker.subscribe(MessageType.selected_event, mock)
        self.message_broker.subscribe(MessageType.mapping, mock)
        calls = [
            call(self.data_manager.active_mapping.get_bus_message()),
            call(InputConfig(type=3, code=0)),
        ]
        with patch.object(self.data_manager, "update_mapping", side_effect=KeyError):
            self.controller.set_event_as_analog(False)
            mock.assert_has_calls(calls, any_order=False)

    def test_update_mapping_type_will_ask_user_when_output_symbol_is_set(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        request: UserConfirmRequest = None

        def f(r: UserConfirmRequest):
            nonlocal request
            request = r

        self.message_broker.subscribe(MessageType.user_confirm_request, f)
        self.controller.update_mapping(mapping_type="analog")
        self.assertIn('This will remove "a" from the text input', request.msg)

    def test_update_mapping_type_will_notify_user_to_recorde_analog_input(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(output_symbol=None)
        request: UserConfirmRequest = None

        def f(r: UserConfirmRequest):
            nonlocal request
            request = r

        self.message_broker.subscribe(MessageType.user_confirm_request, f)
        self.controller.update_mapping(mapping_type="analog")
        self.assertIn("You need to record an analog input.", request.msg)

    def test_update_mapping_type_will_tell_user_which_input_is_used_as_analog(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (2, 1, 1)),
            output_symbol=None,
        )
        request: UserConfirmRequest = None

        def f(r: UserConfirmRequest):
            nonlocal request
            request = r

        self.message_broker.subscribe(MessageType.user_confirm_request, f)
        self.controller.update_mapping(mapping_type="analog")
        self.assertIn('The input "Y Down 1" will be used as analog input.', request.msg)

    def test_update_mapping_type_will_will_autoconfigure_the_input(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (2, 1, 1)),
            output_symbol=None,
        )

        self.message_broker.subscribe(
            MessageType.user_confirm_request, lambda r: r.respond(True)
        )
        with patch.object(self.data_manager, "update_mapping") as mock:
            self.controller.update_mapping(mapping_type="analog")
            mock.assert_called_once_with(
                mapping_type="analog",
                output_symbol=None,
                input_combination=InputCombination(
                    InputCombination.from_tuples((1, 3), (2, 1))
                ),
            )

    def test_update_mapping_type_will_abort_when_user_denys(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )

        self.message_broker.subscribe(
            MessageType.user_confirm_request, lambda r: r.respond(False)
        )
        with patch.object(self.data_manager, "update_mapping") as mock:
            self.controller.update_mapping(mapping_type="analog")
            mock.assert_not_called()

        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (2, 1)),
            output_symbol=None,
            mapping_type="analog",
        )
        with patch.object(self.data_manager, "update_mapping") as mock:
            self.controller.update_mapping(mapping_type="key_macro")
            mock.assert_not_called()

    def test_update_mapping_type_will_delete_output_symbol_when_user_confirms(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )

        self.message_broker.subscribe(
            MessageType.user_confirm_request, lambda r: r.respond(True)
        )
        with patch.object(self.data_manager, "update_mapping") as mock:
            self.controller.update_mapping(mapping_type="analog")
            mock.assert_called_once_with(mapping_type="analog", output_symbol=None)

    def test_update_mapping_will_ask_user_to_set_trigger_threshold(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (2, 1)),
            output_symbol=None,
            mapping_type="analog",
        )
        request: UserConfirmRequest = None

        def f(r: UserConfirmRequest):
            nonlocal request
            request = r

        self.message_broker.subscribe(MessageType.user_confirm_request, f)
        self.controller.update_mapping(mapping_type="key_macro")
        self.assertIn('and set a "Trigger Threshold" for "Y".', request.msg)

    def test_update_mapping_update_to_analog_without_asking(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (2, 1)),
            output_symbol=None,
        )
        mock = MagicMock()
        self.message_broker.subscribe(MessageType.user_confirm_request, mock)
        self.controller.update_mapping(mapping_type="analog")
        mock.assert_not_called()

    def test_update_mapping_update_to_key_macro_without_asking(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (2, 1, 1)),
            mapping_type="analog",
            output_symbol=None,
        )
        mock = MagicMock()
        self.message_broker.subscribe(MessageType.user_confirm_request, mock)
        self.controller.update_mapping(mapping_type="key_macro")
        mock.assert_not_called()

    def test_update_mapping_will_remove_output_type_and_code(self):
        prepare_presets()
        self.data_manager.load_group("Foo Device 2")
        self.data_manager.load_preset("preset2")
        self.data_manager.load_mapping(
            InputCombination(InputCombination.from_tuples((1, 3)))
        )
        self.data_manager.update_mapping(
            input_combination=InputCombination.from_tuples((1, 3), (2, 1)),
            output_symbol=None,
            mapping_type="analog",
        )
        self.message_broker.subscribe(
            MessageType.user_confirm_request, lambda r: r.respond(True)
        )
        with patch.object(self.data_manager, "update_mapping") as mock:
            self.controller.update_mapping(mapping_type="key_macro")
            mock.assert_called_once_with(
                mapping_type="key_macro",
                output_type=None,
                output_code=None,
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Reading events from fixtures, making fixtures act like they are sending events."""

from __future__ import annotations

import multiprocessing
from multiprocessing.connection import Connection
from typing import Dict, Tuple

from tests.lib.fixtures import Fixture
from tests.lib.logger import logger

uinput_write_history = []
# for tests that makes the injector create its processes
uinput_write_history_pipe = multiprocessing.Pipe()
pending_events: Dict[Fixture, Tuple[Connection, Connection]] = {}


def read_write_history_pipe():
    """Convert the write history from the pipe to some easier to manage list."""
    history = []
    while uinput_write_history_pipe[0].poll():
        event = uinput_write_history_pipe[0].recv()
        history.append((event.type, event.code, event.value))
    return history


def setup_pipe(fixture: Fixture):
    """Create a pipe that can be used to send events to the reader-service,
    which in turn will be sent to the reader-client
    """
    if pending_events.get(fixture) is None:
        pending_events[fixture] = multiprocessing.Pipe()


def close_pipe(fixture: Fixture):
    if fixture in pending_events:
        pipe1, pipe2 = pending_events[fixture]
        pipe1.close()
        pipe2.close()
        del pending_events[fixture]


def get_events():
    """Get all events written by the injector."""
    return uinput_write_history


def push_event(fixture: Fixture, event, force: bool = False):
    """Make a device act like it is reading events from evdev.

    push_event is like hitting a key on a keyboard for stuff that reads from
    evdev.InputDevice (which is patched in test.py to work that way)

    Parameters
    ----------
    fixture
        For example 'Foo Device'
    event
        The InputEvent to send
    force
        don't check if the event is in fixture.capabilities
    """
    setup_pipe(fixture)
    if not force and (
        not fixture.capabilities.get(event.type)
        or event.code not in fixture.capabilities[event.type]
    ):
        raise AssertionError(f"Fixture {fixture.path} cannot send {event}")
    logger.info("Simulating %s for %s", event, fixture.path)
    pending_events[fixture][0].send(event)


def push_events(fixture: Fixture, events, force=False):
    """Push multiple events."""
    for event in events:
        push_event(fixture, event, force)



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
from typing import Tuple, Dict, Optional

import evdev
from evdev.ecodes import (
    EV_ABS,
    EV_REL,
    REL_WHEEL,
    REL_HWHEEL,
    REL_HWHEEL_HI_RES,
    REL_WHEEL_HI_RES,
)

from inputremapper import exceptions
from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import (
    Mapping,
    WHEEL_SCALING,
    WHEEL_HI_RES_SCALING,
    REL_XY_SCALING,
    DEFAULT_REL_RATE,
)
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.axis_transform import Transformation
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    HandlerEnums,
    InputEventHandler,
)
from inputremapper.input_event import InputEvent, EventActions
from inputremapper.logging.logger import logger


class RelToAbsHandler(MappingHandler):
    """Handler which transforms EV_REL to EV_ABS events.

    High EV_REL input results in high EV_ABS output.
    If no new EV_REL events are seen, the EV_ABS output is set to 0 after
    release_timeout.
    """

    _map_axis: InputConfig  # InputConfig for the relative movement we map
    _output_axis: Tuple[int, int]  # the (type, code) of the output axis
    _transform: Transformation
    _target_absinfo: evdev.AbsInfo

    # infinite loop which centers the output when input stops
    _recenter_loop: Optional[asyncio.Task]
    _moving: asyncio.Event  # event to notify the _recenter_loop

    _previous_event: Optional[InputEvent]
    _observed_rate: float  # input events per second

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        global_uinputs: GlobalUInputs,
        **_,
    ) -> None:
        super().__init__(combination, mapping, global_uinputs)

        # find the input event we are supposed to map. If the input combination is
        # BTN_A + REL_X + BTN_B, then use the value of REL_X for the transformation
        assert (map_axis := combination.find_analog_input_config(type_=EV_REL))
        self._map_axis = map_axis

        assert mapping.output_code is not None
        assert mapping.output_type == EV_ABS
        self._output_axis = (mapping.output_type, mapping.output_code)

        target_uinput = global_uinputs.get_uinput(mapping.target_uinput)
        assert target_uinput is not None
        abs_capabilities = target_uinput.capabilities(absinfo=True)[EV_ABS]
        self._target_absinfo = dict(abs_capabilities)[mapping.output_code]

        max_ = self._get_default_cutoff()
        self._transform = Transformation(
            min_=-max(1, int(max_)),
            max_=max(1, int(max_)),
            deadzone=mapping.deadzone,
            gain=mapping.gain,
            expo=mapping.expo,
        )
        self._moving = asyncio.Event()
        self._recenter_loop = None

        self._previous_event = None
        self._observed_rate = DEFAULT_REL_RATE

    def __str__(self):
        return f"RelToAbsHandler for {self._map_axis}"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def child(self):  # used for logging
        return (
            f"maps to: {self.mapping.get_output_name_constant()} "
            f"{self.mapping.get_output_type_code()} at "
            f"{self.mapping.target_uinput}"
        )

    def _observe_rate(self, event: InputEvent):
        """Watch incoming events and remember how many events appear per second."""
        if self._previous_event is not None:
            delta_time = event.timestamp() - self._previous_event.timestamp()
            if delta_time == 0:
                logger.error("Observed two events with the same timestamp")
                return

            rate = 1 / delta_time
            # mice seem to have a constant rate. wheel events are jaggy and the
            # rate depends on how fast it is turned.
            if rate > self._observed_rate:
                logger.debug("Updating rate to %s", rate)
                self._observed_rate = rate
                self._calculate_cutoff()

        self._previous_event = event

    def _get_default_cutoff(self):
        """Get the cutoff value assuming the default input rate."""
        if self._map_axis.code in [REL_WHEEL, REL_HWHEEL]:
            return self.mapping.rel_to_abs_input_cutoff * WHEEL_SCALING

        if self._map_axis.code in [REL_WHEEL_HI_RES, REL_HWHEEL_HI_RES]:
            return self.mapping.rel_to_abs_input_cutoff * WHEEL_HI_RES_SCALING

        return self.mapping.rel_to_abs_input_cutoff * REL_XY_SCALING

    def _calculate_cutoff(self):
        """Correct the default cutoff with the observed input rate, and set it."""
        # Mice that have very high input rates report low values at the same time.
        # If the rate is high, use a lower cutoff-value. If the rate is low, use a
        # higher cutoff-value.
        cutoff = self._get_default_cutoff()
        cutoff *= DEFAULT_REL_RATE / self._observed_rate

        self._transform.set_range(-max(1, int(cutoff)), max(1, int(cutoff)))

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        self._observe_rate(event)

        if event.input_match_hash != self._map_axis.input_match_hash:
            return False

        if EventActions.recenter in event.actions:
            if self._recenter_loop:
                self._recenter_loop.cancel()
            self._recenter()
            return True

        if not self._recenter_loop or self._recenter_loop.cancelled():
            self._recenter_loop = asyncio.create_task(self._create_recenter_loop())

        self._moving.set()  # notify the _recenter_loop
        try:
            self._write(self._scale_to_target(self._transform(event.value)))
            return True
        except (exceptions.UinputNotAvailable, exceptions.EventNotHandled):
            return False

    def reset(self) -> None:
        if self._recenter_loop:
            self._recenter_loop.cancel()
        self._recenter()

    def _recenter(self) -> None:
        """Recenter the output."""
        self._write(self._scale_to_target(0))

    async def _create_recenter_loop(self) -> None:
        """Coroutine which waits for the input to start moving,
        then waits until the input stops moving, centers the output and repeat.

        Runs forever.
        """
        while True:
            await self._moving.wait()  # input moving started
            while (
                await asyncio.wait(
                    (asyncio.create_task(self._moving.wait()),),
                    timeout=self.mapping.release_timeout,
                )
            )[0]:
                self._moving.clear()  # still moving
            self._recenter()  # input moving stopped

    def _scale_to_target(self, x: float) -> int:
        """Scales a x value between -1 and 1 to an integer between
        target_absinfo.min and target_absinfo.max

        input values above 1 or below -1 are clamped to the extreme values
        """
        factor = (self._target_absinfo.max - self._target_absinfo.min) / 2
        offset = self._target_absinfo.min + factor
        y = factor * x + offset
        if y > offset:
            return int(min(self._target_absinfo.max, y))
        else:
            return int(max(self._target_absinfo.min, y))

    def _write(self, value: int) -> None:
        """Inject."""
        try:
            self.global_uinputs.write(
                (*self._output_axis, value),
                self.mapping.target_uinput,
            )
        except OverflowError:
            # screwed up the calculation of the event value
            logger.error("OverflowError (%s, %s, %s)", *self._output_axis, value)

    def needs_wrapping(self) -> bool:
        return len(self.input_configs) > 1

    def set_sub_handler(self, handler: InputEventHandler) -> None:
        assert False  # cannot have a sub-handler

    def wrap_with(self) -> Dict[InputCombination, HandlerEnums]:
        if self.needs_wrapping():
            return {InputCombination(self.input_configs): HandlerEnums.axisswitch}
        return {}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.validation_errors import (
    MacroError,
    SymbolNotAvailableInTargetError,
)
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import DummyMapping, MacroTestBase


@test_setup
class TestKey(MacroTestBase):
    async def test_1(self):
        macro = Parser.parse("key(1)", self.context, DummyMapping, True)
        one_code = keyboard_layout.get("1")

        await macro.run(self.handler)
        self.assertListEqual(
            self.result,
            [(EV_KEY, one_code, 1), (EV_KEY, one_code, 0)],
        )
        self.assertEqual(self.count_child_macros(macro), 0)

    async def test_named_parameter(self):
        macro = Parser.parse("key(symbol=1)", self.context, DummyMapping, True)
        one_code = keyboard_layout.get("1")

        await macro.run(self.handler)
        self.assertListEqual(
            self.result,
            [(EV_KEY, one_code, 1), (EV_KEY, one_code, 0)],
        )
        self.assertEqual(self.count_child_macros(macro), 0)

    async def test_2(self):
        macro = Parser.parse('key(1).key("KEY_A").key(3)', self.context, DummyMapping)

        await macro.run(self.handler)
        self.assertListEqual(
            self.result,
            [
                (EV_KEY, keyboard_layout.get("1"), 1),
                (EV_KEY, keyboard_layout.get("1"), 0),
                (EV_KEY, keyboard_layout.get("a"), 1),
                (EV_KEY, keyboard_layout.get("a"), 0),
                (EV_KEY, keyboard_layout.get("3"), 1),
                (EV_KEY, keyboard_layout.get("3"), 0),
            ],
        )
        self.assertEqual(self.count_child_macros(macro), 0)

    async def test_key_down_up(self):
        code_a = keyboard_layout.get("a")
        code_b = keyboard_layout.get("b")
        macro = Parser.parse(
            "set(foo, b).key_down($foo).key_up($foo).key_up(a).key_down(a)",
            self.context,
            DummyMapping,
        )
        await macro.run(self.handler)
        self.assertListEqual(
            self.result,
            [
                (EV_KEY, code_b, 1),
                (EV_KEY, code_b, 0),
                (EV_KEY, code_a, 0),
                (EV_KEY, code_a, 1),
            ],
        )

    async def test_raises_error(self):
        Parser.parse("k(1).h(k(a)).k(3)", self.context)  # No error
        self.expect_string_in_error("bracket", "key((1)")
        self.expect_string_in_error("bracket", "k(1))")
        self.assertRaises(MacroError, Parser.parse, "k((1).k)", self.context)
        self.assertRaises(MacroError, Parser.parse, "key(foo=a)", self.context)
        self.assertRaises(
            MacroError, Parser.parse, "key(symbol=a, foo=b)", self.context
        )
        self.assertRaises(MacroError, Parser.parse, "k()", self.context)
        self.assertRaises(MacroError, Parser.parse, "key(invalidkey)", self.context)
        self.assertRaises(MacroError, Parser.parse, 'key("invalidkey")', self.context)
        Parser.parse("key(1)", self.context)  # no error
        self.assertRaises(MacroError, Parser.parse, "k(1, 1)", self.context)
        Parser.parse("key($a)", self.context)  # no error
        self.assertRaises(MacroError, Parser.parse, "key(a)key(b)", self.context)
        # wrong target for BTN_A
        self.assertRaises(
            SymbolNotAvailableInTargetError,
            Parser.parse,
            "key(BTN_A)",
            self.context,
            DummyMapping,
        )


if __name__ == "__main__":
    unittest.main()
